from typing import TYPE_CHECKING, Any, cast

from braces.views import (
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    JSONResponseMixin,
)
from django.contrib import messages
from django.forms import Form
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import Http404, redirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import CreateView, FormView, UpdateView, View
from django.views.generic.base import ContextMixin
from django.views.generic.edit import TemplateResponseMixin
from wildewidgets import (
    BasicHeader,
    Block,
    BreadcrumbBlock,
    CardWidget,
    Datagrid,
    Navbar,
    NavbarMixin,
    WidgetListLayout,
    WidgetStream,
)

from demo.logging import logger
from ldaporm.managers import F

from .forms import (
    LDAPGroupAddForm,
    LDAPGroupAddMemberForm,
    LDAPGroupEditForm,
    LDAPPersonAddForm,
    LDAPPersonEditForm,
    ManagedRoleForm,
    ResetPasswordForm,
)
from .ldap.models import LDAPGroup, LDAPUser, NSRole
from .password.changer import PasswordChanger
from .wildewidgets import (
    BaseBreadcrumbs,
    GroupAddFormWidget,
    GroupAddMemberModalWidget,
    GroupConfigurationWidget,
    GroupDataDetailWidget,
    GroupMembershipTableWidget,
    GroupTableWidget,
    RoleTableWidget,
    Sidebar,
    UserAddFormWidget,
    UserConfigurationWidget,
    UserDataDetailWidget,
    UserRoleWidget,
    UserTableWidget,
    UserVerifyPasswordWidget,
    WildewidgetsMixin,
)

if TYPE_CHECKING:
    from ldaporm.managers import LdapManager

# -----------------------------------------------------------------------------
# User views
# -----------------------------------------------------------------------------


class UserDetailView(
    FormValidMessageMixin,
    FormInvalidMessageMixin,
    WildewidgetsMixin,
    NavbarMixin,
    UpdateView,
):
    """
    View for displaying and editing person details from LDAP.

    This view provides a comprehensive interface for viewing and editing
    information about a :py:class:`~demo.core.ldap.models.LDAPUser`.  It
    includes various widgets for displaying different aspects of the user's
    information and configuration.

    Attributes:
        form_class: The form class for editing person details
        slug_field: The field to use for looking up the person
        slug_url_kwarg: The URL kwarg that contains the person ID
        navbar_class: The class to use for the navigation sidebar
        menu_item: The menu item to highlight in the navigation

    """

    form_class: type[Form] = LDAPPersonEditForm
    slug_field: str = "uid"
    slug_url_kwarg: str = "uid"
    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Users"

    def get_queryset(self) -> F:
        """
        Get the queryset for the user detail view.
        """
        return cast("LdapManager", LDAPUser.objects)

    def get_content(self) -> WidgetListLayout:
        """
        Build the wildewidgets content layout for the person detail view.

        This method creates a layout with various widgets displaying different
        aspects of the person's configuration, including user information,
        password changing and validation and their roles.

        Returns:
            A layout containing all the person detail widgets

        """
        layout = WidgetListLayout(
            self.object.full_name, sidebar_title="Info", sidebar_width=4
        )
        layout.add_widget(UserConfigurationWidget(self.object))
        layout.add_widget(UserRoleWidget(self.object))
        layout.add_widget(UserVerifyPasswordWidget(self.object))
        layout.add_sidebar_bare_widget(UserDataDetailWidget(self.object))
        layout.add_sidebar_form_button(
            "Reset Password",
            reverse("core:user--reset-password", kwargs={"uid": self.object.uid}),
            data={"uid": self.object.uid},
        )
        return layout

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.

        Returns:
            Breadcrumb navigation showing the path to this person

        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Users", reverse("core:user--list"))
        breadcrumbs.add_breadcrumb(self.object.full_name)
        return breadcrumbs

    def get_success_url(self) -> str:
        """
        Get the URL to redirect to after successful form submission.

        Returns:
            URL to redirect to

        """
        return reverse("core:user--detail", kwargs={"uid": self.object.uid})

    def get_form_invalid_message(self) -> str:
        """
        Get the message to display when form submission is invalid.

        Returns:
            Error message for invalid form submission

        """
        return (
            f"There was a problem updating the {self.object.full_name} "
            f"(uid={self.object.uid}). Please see the errors below."
        )

    def get_form_valid_message(self) -> str:
        """
        Get the message to display when form submission is valid.

        Returns:
            str: Success message for valid form submission

        """
        return f'Updated "{self.object.full_name}" (uid={self.object.uid})'

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.

        This method saves the form data and adds a delay to allow LDAP
        changes to propagate.

        Args:
            form: The validated form

        Returns:
            Redirect to success URL

        """
        user = cast("LdapManager", LDAPUser.objects).get(uid=form.cleaned_data["uid"])
        user.employee_type = form.cleaned_data["employee_type"]
        user.login_shell = form.cleaned_data["login_shell"]
        user.home_phone = form.cleaned_data["home_phone"]
        user.mobile = form.cleaned_data["mobile"]
        user.save()
        return redirect(self.get_success_url())


class UserAddView(
    WildewidgetsMixin,
    NavbarMixin,
    CreateView,
):
    """
    View for adding a new user.
    """

    form_class: type[Form] = LDAPPersonAddForm
    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Users"

    def get_content(self) -> WidgetListLayout:
        """
        Get the content for the user add view.
        """
        return UserAddFormWidget()

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.
        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Users", reverse("core:user--list"))
        breadcrumbs.add_breadcrumb("Add User")
        return breadcrumbs

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.
        """
        user = LDAPUser(
            uid=form.cleaned_data["uid"],
            full_name=form.cleaned_data["full_name"],
            first_name=form.cleaned_data["first_name"],
            last_name=form.cleaned_data["last_name"],
            employee_number=form.cleaned_data["employee_number"],
            employee_type=form.cleaned_data["employee_type"],
            mail=form.cleaned_data["mail"],
            login_shell=form.cleaned_data["login_shell"],
            home_phone=form.cleaned_data["home_phone"],
            mobile=form.cleaned_data["mobile"],
            uid_number=form.cleaned_data["uid_number"],
            gid_number=form.cleaned_data["gid_number"],
            home_directory=form.cleaned_data["home_directory"],
        )
        user.save()
        messages.success(self.request, f'Added "{user.full_name}" (uid={user.uid})')
        return redirect(reverse("core:user--detail", kwargs={"uid": user.uid}))

    def form_invalid(self, form: Form) -> HttpResponse:
        """
        Handle invalid form submission.
        """
        messages.error(self.request, "There was a problem adding the user.")
        # add the list of errors to the message
        for field, errors in form.errors.items():
            messages.error(self.request, f"{field}: {errors}")
        for error in form.non_field_errors():
            messages.error(self.request, error)
        return super().form_invalid(form)


class UserListView(
    WildewidgetsMixin, NavbarMixin, ContextMixin, TemplateResponseMixin, View
):
    """
    View for displaying a list of users.

    This view provides a dataTable.js table of all
    :py:class:`~demo.core.ldap.models.LdapUser` objects.

    Attributes:
        navbar_class: The class to use for the navigation sidebar
        menu_item: The menu item to highlight in the navigation

    """

    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Users"

    def get_queryset(self) -> F:
        """
        Get the queryset for the user list view.
        """
        return cast("LdapManager", LDAPUser.objects).all()

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # noqa: ARG002
        """
        Handle GET requests for the user list view.

        Args:
            request: The HTTP request
            *args: Additional positional arguments

        Keyword Args:
            **kwargs: Additional keyword arguments

        Returns:
            HTTP response with the user list page

        """
        return self.render_to_response(self.get_context_data())

    def get_content(self) -> UserTableWidget:
        """
        Get the content for the user list view.
        """
        return UserTableWidget()

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.

        Returns:
            Breadcrumb navigation showing the path to this person

        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Users")
        return breadcrumbs


class UserResetPasswordView(
    FormInvalidMessageMixin,
    WildewidgetsMixin,
    NavbarMixin,
    FormView,
):
    """
    The view that handles the POST after the user has submitted the form from
    the "Reset Password" button on the user detail page.
    """

    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Search"
    form_class: type[Form] = ResetPasswordForm
    form_invalid_message: str = "No username provided when resetting password."

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Actually reset the user's password by using the
        :py:class:`demo.core.password.changer.PasswordChanger`
        class, which will reset passwords in CAP, LDAP, and AD.

        Args:
            form: The password change form

        Returns:
            A response with the new password and phonetic spelling of the password.

        """
        self.uid = form.cleaned_data["uid"]
        pc = PasswordChanger()
        self.password, self.phonetic_strings = pc.set_random_password(self.uid)
        logger.info("user.password_reset.success", user_uid=self.uid)
        return self.render_to_response(self.get_context_data())

    def form_invalid(self, form: Form) -> HttpResponseRedirect:  # noqa: ARG002
        """
        Something was wrong with our password reset form. Log a warning and
        redirect the user back to the search page.

        Args:
            form: The password reset form

        Returns:
            A redirect to the search page.

        """
        logger.warning("user.password_reset.no_user_provided")
        return redirect(reverse("core:user--list"))

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Build the breadcrumbs for this view.

        Returns:
            The completed breadcrumbs widget

        """
        user = cast("LdapManager", LDAPUser.objects).get(uid=self.uid)
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("User")
        breadcrumbs.add_breadcrumb(
            user.full_name,
            reverse("core:user--detail", kwargs={"uid": self.uid}),
        )
        breadcrumbs.add_breadcrumb("Reset Password")
        return breadcrumbs

    def get_content(self) -> WidgetListLayout:
        """
        Supply the wildewidgets content for this view.   ``get_content`` is
        a method that is called by the ``WildewidgetsMixin`` to get the
        content for the view.  This part of the view is responsible for
        creating the widgets that will be displayed on the page after
        the password has been reset.

        It shows the password, the phonetic spelling of the password, and
        some information about the user.

        Returns:
            The set of widgets that will be displayed on the page.

        """
        user = cast("LdapManager", LDAPUser.objects).get(uid=self.uid)
        grid = Datagrid()
        grid.add_item(title="Employee Number", content=str(user.employee_number))
        grid.add_item(title="Username", content=user.uid)
        grid.add_item(title="E-mail", content=user.mail[0])
        if user.home_phone:
            grid.add_item(title="Phone", content=user.home_phone)
        elif user.mobile:
            grid.add_item(title="Phone", content=f"{user.mobile} [mobile]")
        else:
            grid.add_item(title="Phone", content="No phone number")
        grid.add_item(
            title="Password",
            content=Block(self.password, tag="mark", css_class="font-monospace"),
        )
        phonetics = Block(
            tag="div",
            css_class="list-group",
        )
        for phonetic_string in self.phonetic_strings:
            phonetics.add_block(
                Block(phonetic_string, tag="li", css_class="list-group-item p-2")
            )
        block = Block(
            grid,
            BasicHeader(header_level=3, header_text="Phonetic spelling"),
            phonetics,
        )
        return WidgetStream(
            widgets=[
                BasicHeader(
                    header_text=f"Password Reset for {user.full_name}",
                ),
                CardWidget(widget=block, css_class=CardWidget.css_class + " shadow"),
            ],
            css_class=WidgetStream.css_class + " w-50",
        )


@method_decorator(csrf_exempt, name="dispatch")
class VerifyPasswordAPI(JSONResponseMixin, View):
    """
    Verify the password for a user.  This means try binding each of our
    ldap servers with the provided username and password: CAP, LDAP, and AD.

    If ``settings.TRY_AD`` is False, then we will not try to bind to AD.
    """

    def post(self, request: HttpRequest, *args, **kwargs):
        """
        Handle the POST from the "Verify Password" button on the user detail page.

        Args:
            request: the request object
            *args: the positional arguments used by the view

        Keyword Args:
            **kwargs: the keyword arguments used by the view

        Returns:
            The JSON response with the status of the password verification.

        """
        return self.get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:  # noqa: ARG002
        context: dict[str, Any] = {}
        uid = self.request.POST.get("uid", None)
        password = self.request.POST.get("verify_password_value", None)
        if password is None or uid is None:
            context["status"] = "bad-form"
            return context
        pc = PasswordChanger()
        result = pc.verify_password(uid, password)
        context["result"] = "success" if result else "failure"
        return context

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # noqa: ARG002
        return self.render_json_response(self.get_context_data())


class UserRolesUpdateView(
    FormValidMessageMixin,
    FormView,
):
    """
    View for updating the roles for a user.
    """

    form_class: type[Form] = ManagedRoleForm
    slug_field: str = "uid"
    slug_url_kwarg: str = "uid"

    def get_queryset(self) -> F:
        """
        Get the queryset for the user detail view.
        """
        return cast("LdapManager", LDAPUser.objects)

    def get_form(self, form_class: type[Form] | None = None) -> ManagedRoleForm:  # noqa: ARG002
        """
        Get the form instance for managing roles.

        This method retrieves the user from LDAP and initializes the form
        with the user's current roles.

        Args:
            form_class: Unused parameter (overridden from parent class)

        Returns:
            The initialized form for managing roles

        """
        self.uid = self.request.POST["uid"]
        self.object = cast("LdapManager", LDAPUser.objects).get(uid=self.uid)
        return ManagedRoleForm(self.object, self.request.POST)

    def get_success_url(self) -> str:
        """
        Redirect to the user detail page after updating the roles.
        """
        return reverse("core:user--detail", kwargs={"uid": self.object.uid})

    def get_form_valid_message(self) -> str:
        """
        Get the message to display when form submission is valid.
        """
        return f"Updated roles for {self.object.full_name} (uid={self.object.uid})"

    def form_invalid(self, form: Form) -> HttpResponse:
        """
        Handle invalid form submission.

        This method creates a detailed error message listing all form errors
        and logs the failure.

        Args:
            form: The invalid form

        Returns:
            Redirect to success URL with error message

        """
        # Build a detailed error message
        error_messages = []
        for field_name, errors in form.errors.items():
            field_label = (
                form.fields[field_name].label
                if field_name in form.fields
                else field_name
            )
            error_messages.extend(f"{field_label}: {error}" for error in errors)

        detailed_message = (
            f"There was a problem updating the roles for {self.object.full_name} "
            f"(uid={self.object.uid}). Errors: {'; '.join(error_messages)}"
        )

        # Add the error message to the messages framework
        from django.contrib import messages

        messages.error(self.request, detailed_message)

        return redirect(self.get_success_url())

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.
        """
        self.object.nsroledn = form.cleaned_data["nsroledn"]
        self.object.save()
        return super().form_valid(form)


# -----------------------------------------------------------------------------
# Group views
# -----------------------------------------------------------------------------


class GroupDetailView(
    FormValidMessageMixin,
    FormInvalidMessageMixin,
    WildewidgetsMixin,
    NavbarMixin,
    UpdateView,
):
    """
    View for displaying and editing posix group details from LDAP.

    This view provides a comprehensive interface for viewing and editing
    information about a :py:class:`~demo.core.ldap.models.LDAPGroup` It
    includes various widgets for displaying different aspects of the group's
    information and configuration.

    Attributes:
        form_class: The form class for editing person details
        slug_field: The field to use for looking up the person
        slug_url_kwarg: The URL kwarg that contains the person ID
        navbar_class: The class to use for the navigation sidebar
        menu_item: The menu item to highlight in the navigation

    """

    form_class: type[Form] = LDAPGroupEditForm
    slug_field: str = "gid_number"
    slug_url_kwarg: str = "gid"
    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Groups"

    def get_queryset(self) -> F:
        """
        Get the queryset for the group detail view.
        """
        return cast("LdapManager", LDAPGroup.objects)

    def get_content(self) -> WidgetListLayout:
        """
        Build the wildewidgets content layout for the person detail view.

        This method creates a layout with various widgets displaying different
        aspects of the groups's configuration, including name, description,
        and group membership.

        Returns:
            A layout containing all the person detail widgets

        """
        layout = WidgetListLayout(self.object.cn, sidebar_title="Info", sidebar_width=4)
        layout.add_widget(GroupConfigurationWidget(self.object))
        layout.add_widget(GroupMembershipTableWidget(self.object))
        layout.add_modal(GroupAddMemberModalWidget(self.object))
        layout.add_sidebar_bare_widget(GroupDataDetailWidget(self.object))
        return layout

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.

        Returns:
            Breadcrumb navigation showing the path to this person

        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Groups", reverse("core:group--list"))
        breadcrumbs.add_breadcrumb(self.object.cn)
        return breadcrumbs

    def get_success_url(self) -> str:
        """
        Get the URL to redirect to after successful form submission.

        Returns:
            URL to redirect to

        """
        return reverse("core:group--detail", kwargs={"gid": self.object.gid_number})

    def get_form_invalid_message(self) -> str:
        """
        Get the message to display when form submission is invalid.

        Returns:
            Error message for invalid form submission

        """
        return (
            f"There was a problem updating the {self.object.cn} group"
            f"(gid_number={self.object.gid_number}). Please see the errors below."
        )

    def get_form_valid_message(self) -> str:
        """
        Get the message to display when form submission is valid.

        Returns:
            str: Success message for valid form submission

        """
        return f'Updated group "{self.object.cn}" (gid_number={self.object.gid_number})'

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.

        This method saves the form data and adds a delay to allow LDAP
        changes to propagate.

        Args:
            form: The validated form

        Returns:
            Redirect to success URL

        """
        return super().form_valid(form)


class GroupAddView(
    WildewidgetsMixin,
    NavbarMixin,
    CreateView,
):
    """
    View for adding a new group.
    """

    form_class: type[Form] = LDAPGroupAddForm
    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Groups"

    def get_content(self) -> GroupAddFormWidget:
        """
        Get the content for the user add view.
        """
        return GroupAddFormWidget()

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.
        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Groups", reverse("core:group--list"))
        breadcrumbs.add_breadcrumb("Add Group")
        return breadcrumbs

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.
        """
        group = LDAPGroup(
            cn=form.cleaned_data["cn"],
            description=form.cleaned_data["description"],
            gid_number=form.cleaned_data["gid_number"],
        )
        group.save()
        messages.success(
            self.request, f'Added "{group.cn}" (gid_number={group.gid_number})'
        )
        return redirect(reverse("core:group--detail", kwargs={"gid": group.gid_number}))

    def form_invalid(self, form: Form) -> HttpResponse:
        """
        Handle invalid form submission.
        """
        messages.error(self.request, "There was a problem adding the group.")
        # add the list of errors to the message
        for field, errors in form.errors.items():
            messages.error(self.request, f"{field}: {errors}")
        for error in form.non_field_errors():
            messages.error(self.request, error)
        return super().form_invalid(form)


class GroupListView(
    WildewidgetsMixin, NavbarMixin, ContextMixin, TemplateResponseMixin, View
):
    """
    View for displaying a list of posix groups.

    This view provides a dataTable.js table of all
    :py:class:`~demo.core.ldap.models.LDAPGroup` objects.

    Attributes:
        navbar_class: The class to use for the navigation sidebar
        menu_item: The menu item to highlight in the navigation

    """

    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Groups"

    def get_queryset(self) -> F:
        """
        Get the queryset for the group list view.
        """
        return cast("LdapManager", LDAPGroup.objects).all()

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # noqa: ARG002
        """
        Handle GET requests for the user list view.
        """
        return self.render_to_response(self.get_context_data())

    def get_content(self) -> GroupTableWidget:
        """
        Get the content for the user list view.
        """
        return GroupTableWidget()

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.

        Returns:
            Breadcrumb navigation showing the path to this person

        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Groups")
        return breadcrumbs


class GroupAddMemberView(
    FormValidMessageMixin,
    WildewidgetsMixin,
    NavbarMixin,
    FormView,
):
    """
    View for adding a member to a group.
    """

    def setup(self, request: HttpRequest, *args, **kwargs) -> None:
        super().setup(request, *args, **kwargs)
        try:
            self.object = cast("LdapManager", LDAPGroup.objects).get(
                gid_number=self.kwargs["gid"]
            )
        except LDAPGroup.DoesNotExist as e:
            raise Http404 from e

    def get_form(
        self,
        form_class: type[Form] | None = None,  # noqa: ARG002
    ) -> type[Form] | HttpResponse:
        """
        Get the form class for the group add member view.
        """
        return LDAPGroupAddMemberForm(self.object, **self.get_form_kwargs())

    def get_success_url(self) -> str:
        """
        Get the URL to redirect to after successful form submission.
        """
        return reverse("core:group--detail", kwargs={"gid": self.object.gid_number})

    def form_valid(self, form: Form) -> HttpResponse:
        """
        Handle valid form submission.
        """
        user = LDAPUser.objects.get(uid=form.cleaned_data["member_uid"])
        if not self.object.member_uids:
            self.object.member_uids = []
        if user.uid not in self.object.member_uids:
            self.object.member_uids.append(user.uid)
            self.object.save()
        return super().form_valid(form)

    def form_invalid(self, form: Form) -> HttpResponse:
        """
        Handle invalid form submission.
        """
        message = (
            f"There was a problem adding {self.request.POST['member_uid']} to "
            f'group "{self.object.cn}": {form.errors}'
        )
        # add the list of errors to the message
        for field, errors in form.errors.items():
            message += f"<br>{field}: {errors}"
        for error in form.non_field_errors():
            message += f"<br>{error}"
        messages.add_message(self.request, messages.ERROR, message)
        return redirect(self.get_success_url())

    def get_form_valid_message(self) -> str:
        """
        Get the message to display when form submission is valid.
        """
        return f'Added {self.request.POST["member_uid"]} to group "{self.object.cn}"'


class GroupRemoveMemberView(WildewidgetsMixin, NavbarMixin, View):
    """
    View for removing a member from a group.
    """

    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Groups"

    def post(self, request: HttpRequest, *args, **kwargs) -> HttpResponseRedirect:  # noqa: ARG002
        """
        Remove a member from a group.

        Args:
            request: the request object
            *args: the positional arguments used by the view

        Keyword Args:
            **kwargs: the keyword arguments used by the view

        Returns:
            A redirect to the group detail page.

        """
        try:
            user = LDAPUser.objects.get(uid=kwargs["member_uid"])
        except LDAPUser.DoesNotExist:
            return HttpResponse(status=404)
        try:
            group = LDAPGroup.objects.get(uid=kwargs["gid"])
        except LDAPGroup.DoesNotExist:
            return HttpResponse(status=404)
        group.memberUid.remove(user.uid)
        group.save()
        return redirect(reverse("core:group--detail", kwargs={"gid": group.gid_number}))


# -----------------------------------------------------------------------------
# Role views
# -----------------------------------------------------------------------------


class RoleListView(
    WildewidgetsMixin, NavbarMixin, ContextMixin, TemplateResponseMixin, View
):
    """
    View for displaying a list of roles.
    """

    navbar_class: type[Navbar] = Sidebar
    menu_item: str = "Roles"

    def get_queryset(self) -> F:
        """
        Get the queryset for the role list view.
        """
        return cast("LdapManager", NSRole.objects).all()

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # noqa: ARG002
        """
        Handle GET requests for the role list view.
        """
        return self.render_to_response(self.get_context_data())

    def get_content(self) -> RoleTableWidget:
        """
        Get the content for the role list view.
        """
        return RoleTableWidget()

    def get_breadcrumbs(self) -> BreadcrumbBlock:
        """
        Get the breadcrumb navigation for this view.

        Returns:
            Breadcrumb navigation showing the path to this role table

        """
        breadcrumbs = BaseBreadcrumbs()
        breadcrumbs.add_breadcrumb("Roles")
        return breadcrumbs



from demo.core.ldap.models import LDAPGroup, LDAPUser, NSRole
from ldaporm.restframework import LdapModelSerializer


class LDAPUserSerializer(LdapModelSerializer):
    """Serializer for LDAPUser model."""

    class Meta:
        model = LDAPUser
        fields = (
            "uid",
            "full_name",
            "last_name",
            "first_name",
            "mail",
            "uid_number",
            "gid_number",
            "home_directory",
            "login_shell",
            "employee_number",
            "room_number",
            "home_phone",
            "mobile",
            "nsroledn",
            "nsrole",
        )


class LDAPGroupSerializer(LdapModelSerializer):
    """Serializer for LDAPGroup model."""

    class Meta:
        model = LDAPGroup
        fields = (
            "cn",
            "gid_number",
            "member_uids",
            "description",
        )


class NSRoleSerializer(LdapModelSerializer):
    """Serializer for NSRole model."""

    class Meta:
        model = NSRole
        fields = (
            "cn",
            "description",
        )



"""
LDAP ORM model options and metadata.

This module provides the Options class for managing LDAP model metadata, including
field mappings, LDAP server configuration, and model attributes.
"""

from bisect import bisect
from typing import TYPE_CHECKING, cast

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.utils.functional import cached_property
from django.utils.text import camel_case_to_spaces, format_lazy
from django.utils.translation import override

from .fields import CharListField
from .managers import LdapManager

if TYPE_CHECKING:
    from .fields import Field
    from .models import Model

#: The default attributes for the Options class.
DEFAULT_NAMES = (
    "ldap_server",
    "ldap_options",
    "manager_class",
    "basedn",
    "objectclass",
    "extra_objectclasses",
    "verbose_name",
    "verbose_name_plural",
    "ordering",
    "permissions",
    "default_permissions",
    "password_attribute",
    "userid_attribute",
)


class Options:
    """
    Options class for LDAP model metadata and configuration.

    This class manages all the metadata for an LDAP model, including field
    mappings, LDAP server configuration, and model attributes. It provides
    Django-like model options interface for LDAP ORM models.

    This gets instantiated by parsing the ``Meta`` class for the model, and is
    available as ``model._meta`` on the model class.

    If you are subclassing another model, the ``Meta`` classes will be merged in
    MRO (Method Resolution Order) for the subclass.  This means that the
    ``Meta`` class for the subclass will have all the options from the parent
    class, plus any options that are defined or overridden in the subclass.

    Args:
        meta: The Meta class from the model definition.

    """

    def __init__(self, meta) -> None:
        # LDAP related
        #: The key into ``settings.LDAP_SERVERS`` setting that this model uses.
        self.ldap_server: str = "default"
        #: A list of options to pass to the LDAP server.  The only current option is
        #: ``paged_search`` which will enable paged searches.
        self.ldap_options: list[str] = []
        #: The default manager class to use for this model.  This is really only
        #: for internal use
        self.manager_class: type[LdapManager] = LdapManager
        #: The base DN for this model.
        self.basedn: str | None = None
        #: The objectclass for this model.  This will be automatically added to any
        #: search filters to eliminate objects that are not of this type.
        self.objectclass: str | None = None
        #: Extra objectclasses to add to this model when we are creating new records
        #: only. :py:attr:`objectclass` is always added to the object.
        self.extra_objectclasses: list[str] = []
        #: The attribute to use for the userid.  This is used to identify the user
        #: when we are searching for them.  If this is not a user model, this
        #: will be ignored.
        self.userid_attribute: str = "uid"
        #: The attribute to use for the password.  This is used to store the
        #: password for the user.  If this is not a user model, this will be
        #: ignored.
        self.password_attribute: str | None = None

        # other
        #: The verbose name for this model.
        self.verbose_name: str | None = None
        #: The verbose name plural for this model.
        self.verbose_name_plural: str | None = None
        #: The default ordering for this model.  This is a list of field names to order
        #: by.  The fields can be prefixed with ``-`` to order in descending order.
        self.ordering: list[str] = []
        #: The default permissions for this model.  This is a tuple of the
        #: permissions that are applied to the model by default.  This is here
        #: really just to fool Django's ModelForm.
        self.default_permissions: tuple[str, ...] = ("add", "change", "delete", "view")
        #: The permissions for this model.  This is a list of the permissions
        #: that are applied to the model.  This is here really just to fool
        #: Django's ModelForm.
        self.permissions: list[str] = []

        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.model_name: str | None = None
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.object_name: str | None = None
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.meta = meta
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.  It will be
        #: set to the Field with ``primary_key=True`` on the model.
        self.pk: Field | None = None
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.concrete_model: Model | None = None
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.base_manager: LdapManager | None = None
        #: This is set up by the :py:class:`~ldaporm.models.LdapModelBase``
        #: metaclass.  It is not intended to be set by the user.
        self.local_fields: list[Field] = []

        #: Unused.  This is here really just to fool Django's ModelForm.
        self.local_many_to_many: list[Field] = []
        #: Unused.  This is here really just to fool Django's ModelForm.
        self.private_fields: list[Field] = []
        #: Unused.  This is here really just to fool Django's ModelForm.
        self.many_to_many: list[Field] = []

    @property
    def label(self) -> str:
        """
        Get the model label (object name).

        Returns:
            The model's object name.

        """
        return cast("str", self.object_name)

    @property
    def label_lower(self) -> str:
        """
        Get the lowercase model label.

        Returns:
            The lowercase model name.

        """
        return cast("str", self.model_name)

    @property
    def verbose_name_raw(self) -> str:
        """
        Return the untranslated verbose name.

        Returns:
            The untranslated verbose name.

        """
        with override(None):
            return str(self.verbose_name)

    def contribute_to_class(self, cls: type["Model"], name: str) -> None:  # noqa: ARG002
        """
        Used by the :py:class:`~ldaporm.models.LdapModelBase`` metaclass to
        add this :py:class:`Options` instance to a model class.

        Args:
            cls: The model class to contribute to.
            name: The name of the options attribute.

        """
        cls._meta = self
        self.model = cls
        # First, construct the default values for these options.
        self.object_name = cls.__name__
        self.model_name = self.object_name.lower()
        self.verbose_name = camel_case_to_spaces(self.object_name)

        # Next, apply any overridden values from 'class Meta'.
        if self.meta:
            meta_attrs = self.meta.__dict__.copy()
            for attr in self.meta.__dict__:
                # NOTE: We can't modify a dictionary's contents while looping
                # over it, so we loop over the *original* dictionary instead.
                if attr.startswith("_"):
                    del meta_attrs[attr]
            for attr_name in DEFAULT_NAMES:
                if attr_name in meta_attrs:
                    setattr(self, attr_name, meta_attrs.pop(attr_name))
                elif hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))

            # verbose_name_plural is a special case because it uses a 's'
            # by default.
            if self.verbose_name_plural is None:
                # format_lazy returns _StrPromise but we need str
                self.verbose_name_plural = format_lazy("{}s", self.verbose_name)  # type: ignore[assignment]

            # Any leftover attributes must be invalid.
            if meta_attrs != {}:
                msg = "'class Meta' got invalid attribute(s): {}".format(
                    ",".join(meta_attrs)
                )
                raise TypeError(msg)
        else:
            # format_lazy returns _StrPromise but we need str
            self.verbose_name_plural = format_lazy("{}s", self.verbose_name)  # type: ignore[assignment]
        del self.meta

    def _prepare(self, model: type["Model"]) -> None:
        """
        Used by the :py:class:`~ldaporm.models.LdapModelBase`` metaclass to
        prepare the model after all fields have been added.

        Args:
            model: The model class to prepare.

        Raises:
            ImproperlyConfigured: If the model doesn't have a primary key or
                has a manually defined objectclass field.

        """
        if self.pk is None:
            msg = f"'{self.object_name}' model doesn't have a primary key"
            raise ImproperlyConfigured(msg)
        # Always make sure we have objectclass in our model, so we can filter by it
        # don't call self.attributes here, because that gets cached
        for f in self._get_fields():
            if f.ldap_attribute == "objectclass":
                msg = (
                    "The objectclass field is defined automatically; don't "
                    f"manually define it on the '{self.object_name}' model"
                )
                raise ImproperlyConfigured(msg)
        objectclass = CharListField("objectclass", editable=False, max_length=255)
        model.add_to_class("objectclass", objectclass)

    def add_field(self, field: "Field") -> None:
        """
        Used by the :py:class:`~ldaporm.models.LdapModelBase`` metaclass to
        add a field to the model.

        Args:
            field: The field to add.

        """
        self.local_fields.insert(bisect(self.local_fields, field), field)
        self.setup_pk(field)

    def setup_pk(self, field: "Field") -> None:
        """
        Used by the :py:class:`~ldaporm.models.LdapModelBase`` metaclass to
        set up the primary key field.

        Args:
            field: The field to check for primary key status.

        """
        if not self.pk and field.primary_key:
            self.pk = field

    def __repr__(self) -> str:
        """
        Return a string representation of the Options instance.

        Returns:
            A string representation including the object name.

        """
        return f"<Options for {self.object_name}>"

    @cached_property
    def fields(self) -> list["Field"]:
        """
        Get all fields for this model.

        Returns:
            A list of all fields.

        """
        return self._get_fields()

    @property
    def concrete_fields(self) -> list["Field"]:
        """
        Get concrete fields for this model (alias for fields).

        This is here to fool Django's :py:class:`~django.forms.ModelForm`.

        Returns:
            A list of all fields.

        """
        return self.fields

    def get_fields(
        self,
        include_parents: bool = True,  # noqa: ARG002
        include_hidden: bool = False,  # noqa: ARG002
    ) -> list["Field"]:
        """
        Get fields for this model.  This is here to fool Django's
        :py:class:`~django.forms.ModelForm`.

        Args:
            include_parents: Whether to include parent fields (unused).
            include_hidden: Whether to include hidden fields (unused).

        Returns:
            A list of all fields.

        """
        return self._get_fields()

    def _get_fields(self) -> list["Field"]:
        """
        Get the local fields for this model.

        Returns:
            A list of local fields.

        """
        return self.local_fields

    @cached_property
    def fields_map(self) -> dict[str, "Field"]:
        """
        Get a mapping of field names to field instances.  This is used by
        the :py:class:`~ldaporm.manager.LdapManager`` to get the field
        instances for a model.

        Returns:
            A dictionary mapping field names to field instances.

        """
        res = {}
        fields = self._get_fields()
        for field in fields:
            res[cast("str", field.name)] = field
        return res

    @cached_property
    def attributes_map(self) -> dict[str, str]:
        """
        Get a mapping of field names to LDAP attribute names.  This is used by
        the :py:class:`~ldaporm.manager.LdapManager`` to map LDAP attribute names
        to :py:class:`~ldaporm.fields.Field` instances for a model.

        Returns:
            A dictionary mapping field names to LDAP attribute names.

        """
        res = {}
        fields = self._get_fields()
        for field in fields:
            res[cast("str", field.name)] = field.ldap_attribute
        return res

    @cached_property
    def attribute_to_field_name_map(self) -> dict[str, str]:
        """
        Get a mapping of LDAP attribute names to field names.  This is used by
        the :py:class:`~ldaporm.manager.LdapManager`` to map LDAP attribute names
        to python field names for a model.

        Returns:
            A dictionary mapping LDAP attribute names to field names.

        """
        return {f.ldap_attribute: cast("str", f.name) for f in self._get_fields()}

    @cached_property
    def attributes(self) -> list[str]:
        """
        Get a list of LDAP attribute names for all fields.  This is used by
        the :py:class:`~ldaporm.manager.LdapManager`` to get the LDAP attribute
        names for a model.

        Returns:
            A list of LDAP attribute names.

        """
        return [f.ldap_attribute for f in self._get_fields()]

    def get_field(self, field_name: str) -> "Field":
        """
        Return a field instance given the name of a forward or reverse field.

        Args:
            field_name: The name of the field to retrieve.

        Returns:
            The field instance.

        Raises:
            FieldDoesNotExist: If no field with the given name exists.

        """
        try:
            # Retrieve field instance by name from cached or just-computed
            # field map.
            return self.fields_map[field_name]
        except KeyError as e:
            msg = f"{self.object_name} has no field named '{field_name}'"
            raise FieldDoesNotExist(msg) from e



import environ

env = environ.Env()

##### General #####
bind = "unix:/tmp/app.sock"
workers = 8
worker_class = "sync"
daemon = False
timeout = 300
worker_tmp_dir = "/tmp"  # noqa: S108
# requires futures module for threads > 1
threads = 1

##### Devel #####
reload = True

##### Logging #####
logconfig_dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "root": {
        "handlers": ["syslog_console"],
        "level": "INFO",
    },
    "loggers": {
        "gunicorn.error": {
            "level": "INFO",
            "handlers": ["syslog_console"],
            "propagate": False,
            "qualname": "gunicorn.error",
        },
        "gunicorn.access": {
            "level": "INFO",
            "handlers": ["access_console"],
            "propagate": False,
            "qualname": "gunicorn.access",
        },
    },
    "handlers": {
        "syslog_console": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "formatter": "syslog",
        },
        "access_console": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "formatter": "access_log",
        },
    },
    "formatters": {
        "syslog": {
            "()": "logging.Formatter",
            "fmt": "SYSLOG %(asctime)s [%(levelname)s] %(name)s: %(message)s",
        },
        "access_log": {
            "()": "logging.Formatter",
            "fmt": "GUNICORN_ACCESS %(asctime)s %(message)s",
        },
    },
}



from .settings import *  # noqa: F403


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "testdb",
    }
}

LDAP_SERVERS = {
    "default": {
        "basedn": "o=example,c=us",  # type: ignore[arg-type, assignment]
        "read": {
            "url": "ldap://localhost:389",  # type: ignore[arg-type, assignment]
            "user": None,
            "password": None,
        },
        "write": {
            "url": "ldap://localhost:389",  # type: ignore[arg-type, assignment]
            "user": None,
            "password": None,
        },
    }
}

# REST Framework settings for testing
REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 10,
    "DEFAULT_FILTER_BACKENDS": [],
    "DEFAULT_ORDERING": [],
}



from django.conf import settings
from django.contrib import admin
from django.urls import include, path
from wildewidgets import WildewidgetDispatch

urlpatterns = [
    path("", include("demo.core.urls", namespace="core")),
    path("api/", include("demo.api.urls", namespace="api")),
    path("accounts/", include("django.contrib.auth.urls")),
    path("admin/", include(admin.site.urls[:2], namespace=admin.site.name)),
    path("wildewidgets_json", WildewidgetDispatch.as_view(), name="wildewidgets_json"),
]


if settings.ENABLE_DEBUG_TOOLBAR:
    import debug_toolbar

    urlpatterns.append(path("__debug__/", include(debug_toolbar.urls)))



import os

os.environ["DB_NAME"] = "fake"
os.environ["DB_USER"] = "fake"
os.environ["DB_PASSWORD"] = "fake"  # noqa: S105
os.environ["DB_HOST"] = "fake"
os.environ["DB_PORT"] = "fake"
os.environ["CACHE"] = "fake"
os.environ["DEVELOPMENT"] = "True"
os.environ["ENABLE_DEBUG_TOOLBAR"] = "True"

from .settings import *  # noqa: F403



# Generated by Django 4.1.2 on 2022-11-09 23:37

from django.db import migrations
from django.core.management import call_command

fixture = 'users'


def load_fixture(apps, schema_editor):
    call_command('loaddata', fixture, app_label='users')


def unload_fixture(apps, schema_editor):
    User = apps.get_model("user", "User")
    User.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ("users", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(load_fixture, reverse_code=unload_fixture),
    ]



from django import template

from demo.core.ldap.models import LDAPUser

register = template.Library()


@register.filter("role_name")
def role_name(value: str) -> str:
    """
    Return the cn of a role.

    Args:
        value: The role value

    Returns:
        The cn of the role.

    """
    return value.split(",")[0].split("=")[1]


@register.filter("derived_roles")
def derived_roles(person: LDAPUser) -> list[str]:
    return [role for role in person.nsrole if role not in person.nsroledn]  # type: ignore[operator]



from django.urls import path

from . import views

app_name: str = "core"

urlpatterns = [
    path("", views.UserListView.as_view(), name="user--list"),
    path("user/<str:uid>/edit/", views.UserDetailView.as_view(), name="user--detail"),
    path("user/add/", views.UserAddView.as_view(), name="user--add"),
    path(
        "users/<str:uid>/roles/",
        views.UserRolesUpdateView.as_view(),
        name="user--roles--update",
    ),
    path(
        "users/<str:uid>/reset-password/",
        views.UserResetPasswordView.as_view(),
        name="user--reset-password",
    ),
    path(
        "users/<str:uid>/verify-password/",
        views.VerifyPasswordAPI.as_view(),
        name="user--verify-password",
    ),
    path("groups/", views.GroupListView.as_view(), name="group--list"),
    path(
        "group/<str:gid>/edit/", views.GroupDetailView.as_view(), name="group--detail"
    ),
    path("group/add/", views.GroupAddView.as_view(), name="group--add"),
    path(
        "group/<str:gid>/add-member/",
        views.GroupAddMemberView.as_view(),
        name="group--user--add",
    ),
    path(
        "group/<str:gid>/remove-member/",
        views.GroupRemoveMemberView.as_view(),
        name="group--user--remove",
    ),
    path("roles/", views.RoleListView.as_view(), name="role--list"),
]



import structlog
from crequest.middleware import CrequestMiddleware

logger = structlog.get_logger("sphinx_hosting_demo")


def request_context_logging_processor(_, __, event_dict):
    """
    Adds extra runtime event info to our log messages based on the current request.

      ``username``: the username of the logged in user, if user is logged in.
      ``remote_ip``: the REMOTE_ADDR address. django-xff will handle properly setting
            this if we're behind a proxy
      ``superuser``: True if the current User is a superuser

    Does not overwrite any event info that's already been set in the logging call.
    """
    request = CrequestMiddleware.get_request()
    if request is not None:
        try:
            # django-xff will set this appropriately to the actual client IP when
            # we are behind a proxy
            client_ip = request.META["REMOTE_ADDR"]
        except AttributeError:
            # Sometimes there will be a current request, but it's not a real
            # request (during tests). If we can't get the real client ip, just
            # use a placeholder.
            client_ip = "fake IP"
        event_dict.setdefault("remote_ip", client_ip)
        event_dict.setdefault("username", request.user.username or "AnonymousUser")
        event_dict.setdefault("superuser", request.user.is_superuser)
    return event_dict


def censor_password_processor(_, __, event_dict):
    """
    Automatically censors any logging context key called "password",
    "password1", or "password2".
    """
    for password_key_name in ("password", "password1", "password2"):
        if password_key_name in event_dict:
            event_dict[password_key_name] = "*CENSORED*"
    return event_dict



import re
from re import Pattern
from typing import TYPE_CHECKING, ClassVar, Final, cast

from demo.core.ldap.models import LDAPUser

if TYPE_CHECKING:
    from ldaporm.managers import LdapManager


class ValidationException(Exception):
    """
    Exception raised when password validation fails.

    This exception includes specific error codes and messages for different
    types of validation failures.

    Args:
        code: The specific error code for this validation failure

    """

    errors: Final[dict[str, str]] = {
        "generic": "Password validation error",
        "length": (
            "Please select a password that is at least ten (10) characters in length."
        ),
        "need-non-alpha": (
            "Please select a password that contains at least one (1) "
            "non-alphabetic character."
        ),
        "need-alpha": (
            "Please select a password that contains at least (1) alphabetic "
            "characters each in upper and lower case."
        ),
        "restricted": (
            "Please select a password that does not contain double "
            'quote ("), plus (+), equal (=) or percent (%) characters.'
        ),
        "forbidden-words": (
            "Please select a password that does not contain your username, first name, "
            "last name, or full middle name."
        ),
        "invalid-current": "Invalid current password.",
        "must-be-different": (
            "Your new password must be different than your current password."
        ),
    }

    def __init__(self, code: str = "generic") -> None:
        self.code = code

    def __str__(self) -> str:
        return self.errors[self.code]


class AbstractValidationRule:
    """
    Abstract base class for password validation rules.

    This class defines the interface for all password validation rules.
    Subclasses must implement the validate method.
    """

    def validate(self, old_password: str, new_password: str, username: str) -> None:
        """
        Validate a password against this rule.

        Args:
            old_password: The user's current password
            new_password: The proposed new password
            username: The username of the user

        Raises:
            ValidationException: If the password fails this validation rule

        """
        raise NotImplementedError


class OldPasswordIsNotEmpty(AbstractValidationRule):
    """
    Validation rule that ensures the old password is not empty.

    This rule checks that the old password contains at least one non-whitespace
    character.
    """

    def validate(
        self,
        old_password: str,
        new_password: str,  # noqa: ARG002
        username: str,  # noqa: ARG002
    ) -> None:
        """
        Validate that the old password is not empty.

        Args:
            old_password: The user's current password
            new_password: The proposed new password (unused)
            username: The username of the user (unused)
            **kwargs: Additional context for validation (unused)

        Raises:
            ValidationException: If the old password is empty or contains only
                whitespace

        """
        if not old_password.strip():
            msg = "length"
            raise ValidationException(msg)


class OldPasswordIsCorrect(AbstractValidationRule):
    """
    Validation rule that ensures the old password is correct.

    This rule authenticates the user with the provided old password to verify it
    is correct.
    """

    def validate(
        self,
        old_password: str,
        new_password: str,  # noqa: ARG002
        username: str,
    ) -> None:
        """
        Validate that the old password is correct.

        Args:
            old_password: The user's current password
            new_password: The proposed new password (unused)
            username: The username of the user
            **kwargs: Additional context for validation (unused)

        Raises:
            ValidationException: If authentication with the old password fails

        """
        if not cast("LdapManager", LDAPUser.objects).authenticate(
            username, old_password
        ):
            msg = "invalid-current"
            raise ValidationException(msg)


class OldAndNewPasswordsDiffer(AbstractValidationRule):
    """
    Validation rule that ensures the new password differs from the old one.

    This rule checks that the new password is not identical to the old password.
    """

    def validate(
        self,
        old_password: str,
        new_password: str,
        username: str,  # noqa: ARG002
    ) -> None:
        """
        Validate that the new password is different from the old password.

        Args:
            old_password: The user's current password
            new_password: The proposed new password
            username: The username of the user (unused)

        Raises:
            ValidationException: If the new password is identical to the old password

        """
        if old_password == new_password:
            msg = "must-be-different"
            raise ValidationException(msg)


class NewPasswordIsNotEmpty(AbstractValidationRule):
    """
    Validation rule that ensures the new password is not empty.

    This rule checks that the new password contains at least one non-whitespace
    character.
    """

    def validate(
        self,
        old_password: str,  # noqa: ARG002
        new_password: str,
        username: str,  # noqa: ARG002
    ) -> None:
        """
        Validate that the new password is not empty.

        Args:
            old_password: The user's current password (unused)
            new_password: The proposed new password
            username: The username of the user (unused)

        Raises:
            ValidationException: If the new password is empty or contains only
                whitespace

        """
        if not new_password.strip():
            msg = "length"
            raise ValidationException(msg)


class NewPasswordLengthIsOK(AbstractValidationRule):
    """
    Validation rule that ensures the new password has an acceptable length.

    This rule checks that the new password is between MIN_LENGTH and MAX_LENGTH
    characters.
    """

    #: The minimum allowable length of a password
    MIN_LENGTH: Final[int] = 10

    def validate(
        self,
        old_password: str,  # noqa: ARG002
        new_password: str,
        username: str,  # noqa: ARG002
    ) -> None:
        """
        Validate that the new password length is within the acceptable range.

        Args:
            old_password: The user's current password (unused)
            new_password: The proposed new password
            username: The username of the user (unused)

        Raises:
            ValidationException: If the new password length is less than MIN_LENGTH

        """
        if len(new_password) < self.MIN_LENGTH:
            msg = "length"
            raise ValidationException(msg)


class NewPasswordContentsAreOk(AbstractValidationRule):
    """
    Validation rule that ensures the new password has acceptable content.

    This rule performs multiple checks on the password content, including:
    - Checking for restricted characters (quote, plus, equal, percent)
    - Ensuring it contains at least one lowercase alphabetic character
    - Ensuring it contains at least one uppercase alphabetic character
    - Ensuring it contains at least one numeric character
    - Ensuring it has characters from all three character classes
    - Ensuring it doesn't contain forbidden words (username, first/last name, etc.)

    """

    #: A regular expression that matches restricted characters in a password
    RESTRICTED_CHARS_RE: Final[Pattern] = re.compile('["%=+]')
    #: A regular expression that matches at least two alphabetic characters
    ALPHA_RE: Final[Pattern] = re.compile("(?=.*[a-zA-Z].*[a-zA-Z])")
    #: A regular expression that matches non-alphabetic characters
    NON_ALPHA_RE: Final[Pattern] = re.compile("[^a-zA-Z]")
    #: A list of regular expressions that match specific character classes
    SPECIFIC_CHARS_RE: Final[list[Pattern]] = [
        re.compile(r"[a-z]"),  # at least 1 lowercase letter
        re.compile(r"[A-Z]"),  # at least 1 uppercase letter
        re.compile(r"[0-9]"),  # at least 1 number
    ]
    #: Minimum number of character classes that must be present in the password
    MIN_CHAR_CLASSES: Final[int] = 3
    #: Min length of a password to check for restricted characters instead
    #: of character classes
    PASSWORD_LENGTH_BREAKPOINT: Final[int] = 20

    def check_forbidden_words(self, password: str, user: LDAPUser) -> None:
        """
        A validation method that throws an error if the user's first name, last
        name, middle name, username, or employee number is incorporated into the
        password by pulling those attributes from the user's person object
        and doing a string comparison.

        Args:
            password: the password input by the user
            user: the user's person object

        Raises:
            PasswordForbiddenWordException: if the password contains the user's
                first name, last name, middle name, username, or Caltech UID

        """
        forbidden_words = [
            user.uid.lower(),
            user.last_name.lower(),
        ]
        # the person may not have an employee number
        if hasattr(user, "employee_number") and user.employee_number:
            forbidden_words.append(str(user.employee_number))
        # the person may not have a first name
        if hasattr(user, "first_name") and user.first_name:
            forbidden_words.append(user.first_name.lower())
        for word in forbidden_words:
            if len(word) > 2:  # noqa: PLR2004
                if word in password.lower():
                    msg = "forbidden-words"
                    raise ValidationException(msg)

    def check_restricted_set(self, password: str) -> None:
        """
        Check for restricted characters in the password.

        Args:
            password: the password input by the user

        Raises:
            ValidationException: if the password contains restricted characters

        """
        if self.RESTRICTED_CHARS_RE.search(password):
            msg = "restricted"
            raise ValidationException(msg)

    def check_character_sets(self, password: str) -> None:
        """
        Check that the password contains at least two alphabetic characters, at
        least one non-alphabetic character, and at least three of the four
        character classes: uppercase letters, lowercase letters, numbers, and
        non-alphanumeric characters.

        Args:
            password: the password input by the user

        Raises:
            ValidationException: The password does not meet the requirements

        """
        # At least two letters
        if not self.ALPHA_RE.search(password):
            msg = "need-alpha"
            raise ValidationException(msg)
        self.check_restricted_set(password)
        # At least one non-letter
        if not self.NON_ALPHA_RE.search(password):
            msg = "need-non-alpha"
            raise ValidationException(msg)

        count = 0
        for pattern in self.SPECIFIC_CHARS_RE:
            if pattern.search(password):
                count = count + 1

        if count < self.MIN_CHAR_CLASSES:
            msg = "char-class"
            raise ValidationException(msg)

    def validate(
        self,
        old_password: str,  # noqa: ARG002
        new_password: str,
        username: str,
    ) -> None:
        """
        Validate the new password.

        If the password is less than 20 characters, we check for restricted
        characters and character classes.  If it is 20 characters or longer, we
        check for restricted sets of characters.

        Args:
            old_password: the user's current password (unused)
            new_password: the new password
            username: the user's username (unused)
            **kwargs: additional keyword arguments

        """
        user = cast("LdapManager", LDAPUser.objects).get(uid=username)
        self.check_forbidden_words(new_password, user)
        self.check_character_sets(new_password)


class AbstractPasswordValidator:
    """
    Abstract base class for password validators.

    A password validator applies a sequence of validation rules to verify
    that a password meets all requirements.

    """

    #: The list of validation rule classes to apply
    rules: ClassVar[list[type[AbstractValidationRule]]]

    def validate(self, old_password: str, new_password: str, username: str) -> None:
        for rule in self.rules:
            rule().validate(old_password, new_password, username)


class PasswordValidator(AbstractPasswordValidator):
    """
    The validator we use when the user knows their old password.
    """

    rules: ClassVar[list[type[AbstractValidationRule]]] = [
        NewPasswordIsNotEmpty,
        OldPasswordIsNotEmpty,
        OldPasswordIsCorrect,
        OldAndNewPasswordsDiffer,
        NewPasswordLengthIsOK,
        NewPasswordContentsAreOk,
    ]


class ForcedResetPasswordValidator(AbstractPasswordValidator):
    """
    The validator we use when someone has used access.caltech Help Desk to force
    reset a user's password.  In this case, we don't know the old password.
    """

    rules: ClassVar[list[type[AbstractValidationRule]]] = [
        NewPasswordIsNotEmpty,
        NewPasswordLengthIsOK,
        NewPasswordContentsAreOk,
    ]


class RandomPasswordValidator(AbstractPasswordValidator):
    """
    The validator we use when we're generating a random password.
    """

    rules: ClassVar[list[type[AbstractValidationRule]]] = [
        NewPasswordIsNotEmpty,
        NewPasswordLengthIsOK,
        NewPasswordContentsAreOk,
    ]



import logging
import logging.config
from typing import Any

import environ
import structlog

from .logging import censor_password_processor, request_context_logging_processor

# The name of our project
# ------------------------------------------------------------------------------
PROJECT_NAME: str = "demo"
HUMAN_PROJECT_NAME: str = "django-ldaporm Demo"

# Load our environment with django-environ
BASE_DIR: environ.Path = environ.Path(__file__) - 2
APPS_DIR: environ.Path = BASE_DIR.path(PROJECT_NAME)

env: environ.Env = environ.Env()

# ==============================================================================

# DEBUG is Django's own config variable, while DEVELOPMENT is ours.
# Use DEVELOPMENT for things that you want to be able to enable/disable in dev
# without altering DEBUG, since that affects lots of other things.
DEBUG: bool = env.bool("DEBUG", default=False)  # type: ignore[arg-type]
DEVELOPMENT: bool = env.bool("DEVELOPMENT", default=False)  # type: ignore[arg-type]
TESTING: bool = env.bool("TESTING", default=False)  # type: ignore[arg-type]
# Set BOOTSTRAP_ALWAYS_MIGRATE to True if you want to always run pending
# migrations on container boot up
BOOTSTRAP_ALWAYS_MIGRATE: bool = env.bool("BOOTSTRAP_ALWAYS_MIGRATE", default=True)  # type: ignore[arg-type]

# GENERAL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#secret-key
# You MUST define a unique DJANGO_SECRET_KEY env var for your app. And be sure
# not to use $ or # in the value.  The default is there for tests ONLY.
SECRET_KEY: str = env(
    "DJANGO_SECRET_KEY",
    default="Z4peQSPAoo8fruA7pRXt2LefJD2lqYvXtgPFxF2hAeUd7NGOpA8fAt3UpnXpJ019",  # type: ignore[arg-type]
)
# https://docs.djangoproject.com/en/3.2/ref/settings/#allowed-hosts
ALLOWED_HOSTS: list[str] = env.list("DJANGO_ALLOWED_HOSTS", default=["*"])  # type: ignore[assignment]

# INTERNATIONALIZATION (i18n) AND LOCALIZATION (l10n)
# Change these values to suit this project.
# https://docs.djangoproject.com/en/3.2/topics/i18n/
# ------------------------------------------------------------------------------
LANGUAGE_CODE: str = "en-us"
TIME_ZONE: str = "America/Los_Angeles"
USE_I18N: bool = False
USE_L10N: bool = False
USE_TZ: bool = True

# DATABASES
# -----------
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases
if TESTING:
    DATABASES: dict[str, dict[str, Any]] = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": BASE_DIR.path("db.sqlite3"),
        }
    }
else:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": env("DB_NAME", default="demo"),  # type: ignore[arg-type]
            "USER": env("DB_USER", default="demo_u"),  # type: ignore[arg-type]
            "PASSWORD": env("DB_PASSWORD", default="password"),  # type: ignore[arg-type]
            "HOST": env("DB_HOST", default="db"),  # type: ignore[arg-type]
            "ATOMIC_REQUESTS": True,
            # This is needed in case the database doesn't have the newer default
            # settings that enable "strict mode".
            "OPTIONS": {
                "sql_mode": "traditional",
            },
        }
    }
DEFAULT_AUTO_FIELD: str = "django.db.models.AutoField"

# CACHES
# ------------------------------------------------------------------------------
# Disable all caching if the optional DISABLE_CACHE env var is True.
if env.bool("DISABLE_CACHE", default=False):  # type: ignore[arg-type]
    CACHES: dict[str, Any] = {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    }
else:
    CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        }
    }

# URLS
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#root-urlconf
# noinspection PyUnresolvedReferences
ROOT_URLCONF: str = f"{PROJECT_NAME}.urls"
# https://docs.djangoproject.com/en/3.2/ref/settings/#wsgi-application
WSGI_APPLICATION: str = f"{PROJECT_NAME}.wsgi.application"

# APPS
# ------------------------------------------------------------------------------
DJANGO_APPS: list[str] = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # 'django.contrib.humanize', # Handy template tags
    "django.contrib.admin",
]
THIRD_PARTY_APPS: list[str] = [
    "django_extensions",
    "sass_processor",
    "crispy_forms",
    "crispy_bootstrap5",
    "academy_theme",
    "wildewidgets",
    "ldaporm",
]
LOCAL_APPS: list[str] = [
    "demo.users",
    "demo.core",
    "demo.api",
]
# https://docs.djangoproject.com/en/3.2/ref/settings/#installed-apps
INSTALLED_APPS: list[str] = DJANGO_APPS + THIRD_PARTY_APPS + LOCAL_APPS

# AUTHENTICATION
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#authentication-backends
AUTHENTICATION_BACKENDS: list[str] = [
    "django.contrib.auth.backends.ModelBackend",
]

# Use our custom User model instead of auth.User, because it's good practice to
# define a custom one at the START.
AUTH_USER_MODEL: str = "users.User"

# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS: list[dict[str, str]] = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"  # noqa: E501
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

# MIDDLEWARE
# ------------------------------------------------------------------------
MIDDLEWARE: list[str] = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "xff.middleware.XForwardedForMiddleware",
    "crequest.middleware.CrequestMiddleware",
]

# STATIC
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#static-root
# noinspection PyUnresolvedReferences
STATIC_ROOT: str = "/static"
# https://docs.djangoproject.com/en/3.2/ref/settings/#static-url
STATIC_URL: str = "/static/"
# https://docs.djangoproject.com/en/3.2/ref/contrib/staticfiles/#std:setting-STATICFILES_DIRS
STATICFILES_DIRS: list[str] = []
# https://docs.djangoproject.com/en/3.2/ref/contrib/staticfiles/#staticfiles-finders
STATICFILES_FINDERS: list[str] = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "sass_processor.finders.CssFinder",
]

# MEDIA
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-file-storage
DEFAULT_FILE_STORAGE: str = "django.core.files.storage.FileSystemStorage"
# https://docs.djangoproject.com/en/3.2/ref/settings/#media-root
# noinspection PyUnresolvedReferences
MEDIA_ROOT: str = "/media"
# https://docs.djangoproject.com/en/3.2/ref/settings/#media-url
MEDIA_URL: str = "/media/"

# TEMPLATES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#templates
TEMPLATES: list[dict[str, Any]] = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {
            # https://docs.djangoproject.com/en/3.2/ref/settings/#template-loaders
            # https://docs.djangoproject.com/en/3.2/ref/templates/api/#loader-types
            # Always use Django's cached templates loader, even in dev with
            # DEBUG==True.  We use RequestLevelTemplateCacheMiddleware in dev to
            # clear that cache at the start of each request, so that the latest
            # template changes get loaded without having to reboot gunicorn.
            "loaders": [
                (
                    "django.template.loaders.cached.Loader",
                    [
                        "django.template.loaders.filesystem.Loader",
                        "django.template.loaders.app_directories.Loader",
                    ],
                ),
            ],
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "academy_theme.context_processors.theme",
            ],
        },
    },
]

# FIXTURES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#fixture-dirs
FIXTURE_DIRS: tuple[str, ...] = (str(APPS_DIR.path("fixtures")),)

# SECURITY
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#secure-proxy-ssl-header
# SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')  # noqa: ERA001
# https://docs.djangoproject.com/en/3.2/ref/settings/#secure-ssl-redirect
# SECURE_SSL_REDIRECT = env.bool('DJANGO_SECURE_SSL_REDIRECT', default=True)  # noqa: E501, ERA001
# https://docs.djangoproject.com/en/3.2/ref/settings/#session-cookie-httponly
SESSION_COOKIE_HTTPONLY: bool = True
# https://docs.djangoproject.com/en/3.2/ref/settings/#session-cookie-secure
SESSION_COOKIE_SECURE: bool = True
# https://docs.djangoproject.com/en/2.0/ref/settings/#session-cookie-name
SESSION_COOKIE_NAME: str = f"{PROJECT_NAME}_session"
# https://docs.djangoproject.com/en/3.2/ref/settings/#csrf-cookie-httponly
CSRF_COOKIE_HTTPONLY: bool = True
# https://docs.djangoproject.com/en/2.0/ref/settings/#csrf-cookie-name
CSRF_COOKIE_NAME: str = f"{PROJECT_NAME}_csrftoken"
# https://docs.djangoproject.com/en/3.0/ref/settings/#csrf-cookie-secure
CSRF_COOKIE_SECURE: bool = True
# https://docs.djangoproject.com/en/3.2/ref/settings/#csrf-trusted-origins
CSRF_TRUSTED_ORIGINS: list[str] = []
# https://docs.djangoproject.com/en/3.2/ref/settings/#secure-browser-xss-filter
SECURE_BROWSER_XSS_FILTER: bool = True
# https://docs.djangoproject.com/en/3.2/ref/settings/#x-frame-options
X_FRAME_OPTIONS: str = "DENY"
# https://docs.djangoproject.com/en/3.2/ref/settings/#session-expire-at-browser-close
# Don't use persistent sessions, since that could lead to a sensitive information leak.
SESSION_EXPIRE_AT_BROWSER_CLOSE: bool = True
# https://docs.djangoproject.com/en/3.2/ref/settings/#session-cookie-age
if not DEBUG:
    # Chrome and Firefox ignore SESSION_EXPIPE_AT_BROWSER_CLOSE
    SESSION_COOKIE_AGE: int = 60 * 120

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-from-email
DEFAULT_FROM_EMAIL: str = env(  # type: ignore[assignment]
    "DJANGO_DEFAULT_FROM_EMAIL",
    default=f"{HUMAN_PROJECT_NAME} <noreply@example.com>",  # type: ignore[arg-type]
)
# https://docs.djangoproject.com/en/3.2/ref/settings/#server-email
SERVER_EMAIL: str = env("DJANGO_SERVER_EMAIL", default=DEFAULT_FROM_EMAIL)  # type: ignore[assignment]
# https://docs.djangoproject.com/en/3.2/ref/settings/#email-subject-prefix
EMAIL_SUBJECT_PREFIX: str = env(
    "DJANGO_EMAIL_SUBJECT_PREFIX",
    default=f"[{HUMAN_PROJECT_NAME}]",  # type: ignore[arg-type]
)
EMAIL_BACKEND: str = "django.core.mail.backends.smtp.EmailBackend"
EMAIL_HOST: str = env("EMAIL_HOST", default="mail")  # type: ignore[arg-type, assignment]
EMAIL_HOST_USER: str | None = env("EMAIL_HOST_USER", default=None)  # type: ignore[arg-type, assignment]
EMAIL_HOST_PASSWORD: str | None = env("EMAIL_HOST_PASSWORD", default=None)  # type: ignore[arg-type, assignment]
EMAIL_USE_TLS: bool = env.bool("EMAIL_USE_TLS", default=False)  # type: ignore[arg-type, assignment]
EMAIL_PORT: int = env.int("EMAIL_PORT", default=1025)  # type: ignore[arg-type, assignment]

# ADMIN
# ------------------------------------------------------------------------------
# Django Admin URL.
ADMIN_URL: str = env("DJANGO_ADMIN_URL", default="admin/")  # type: ignore[arg-type, assignment]
# https://docs.djangoproject.com/en/3.2/ref/settings/#admins
ADMINS: list[tuple[str, str]] = [
    ("django-ldaporm Admins", "django-ldaporm@example.com")
]
# https://docs.djangoproject.com/en/3.2/ref/settings/#managers
MANAGERS = ADMINS

# LOGGING
# ------------------------------------------------------------------------------
# Use structlog to ease the difficulty of adding context to log messages
# See https://structlog.readthedocs.io/en/stable/index.html
structlog.configure(
    processors=[
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        request_context_logging_processor,
        censor_password_processor,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=False,
)

pre_chain = [
    structlog.processors.StackInfoRenderer(),
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.processors.TimeStamper(fmt="iso"),
]

# Build our custom logging config.
LOGGING_CONFIG: dict[str, Any] | None = None
LOGGING: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "root": {
        "handlers": ["structlog_console"],
        "level": "INFO",
    },
    "loggers": {
        "django": {
            "handlers": ["structlog_console"],
            "level": "INFO",
            "propagate": False,
        },
        "django.security.DisallowedHost": {
            # We don't care about attempts to access the site with a spoofed
            # HTTP-HOST header. It clutters the logs.
            "handlers": ["null"],
            "propagate": False,
        },
        "qinspect": {
            # This is the QueryInspect logger. We always configure it (to
            # simplify the logging setup code), but it doesn't get used unless
            # we turn on QueryInspect.
            "handlers": ["structlog_console"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
    "handlers": {
        "structlog_console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "structlog",
        },
        "null": {
            "level": "DEBUG",
            "class": "logging.NullHandler",
        },
    },
    "formatters": {
        # Set up a special formatter for our structlog output
        "structlog": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.dev.ConsoleRenderer(repr_native_str=False),
            "foreign_pre_chain": pre_chain,
            # Prefix our logs with SYSLOG to make them easier to either grep in or out
            "format": "SYSLOG %(message)s",
        },
    },
}
logging.config.dictConfig(LOGGING)


# crispy-forms
# ------------------------------------------------------------------------------
CRISPY_ALLOWED_TEMPLATE_PACKS: str = "bootstrap5"
CRISPY_TEMPLATE_PACK: str = "bootstrap5"


# django-theme-academy
# ------------------------------------------------------------------------------
ACADEMY_THEME_SETTINGS: dict[str, Any] = {
    # Header
    "APPLE_TOUCH_ICON": "core/images/apple-touch-icon.png",
    "FAVICON_32": "core/images/favicon-32x32.png",
    "FAVICON_16": "core/images/favicon-16x16.png",
    "FAVICON": "core/images/favicon.ico",
    "SITE_WEBMANIFEST": "core/images/site.webmanifest",
    # Footer
    "ORGANIZATION_LINK": "https://github.com/caltechads/django-ldaporm",
    "ORGANIZATION_NAME": "django-ldaporm",
    "ORGANIZATION_ADDRESS": "123 Main Street, Everytown, ST",
    "COPYRIGHT_ORGANIZATION": "Django LDAPORM",
    "FOOTER_LINKS": [],
}

# django-wildewidgets
# ------------------------------------------------------------------------------
WILDEWIDGETS_DATETIME_FORMAT: str = "%Y-%m-%d %H:%M %Z"

# django-xff
# ------------------------------------------------------------------------------
XFF_TRUSTED_PROXY_DEPTH: int = env.int("XFF_TRUSTED_PROXY_DEPTH", default=1)  # type: ignore[arg-type]
XFF_HEADER_REQUIRED: bool = env.bool("XFF_HEADER_REQUIRED", default=False)  # type: ignore[arg-type]


# django-debug-toolbar
# ------------------------------------------------------------------------------
# We don't enable the debug toolbar unless DEVELOPMENT is also True.
ENABLE_DEBUG_TOOLBAR = DEVELOPMENT and env.bool("ENABLE_DEBUG_TOOLBAR", default=False)  # type: ignore[arg-type]
if ENABLE_DEBUG_TOOLBAR:
    # https://django-debug-toolbar.readthedocs.io/en/latest/installation.html#prerequisites
    INSTALLED_APPS += ["debug_toolbar"]
    # https://django-debug-toolbar.readthedocs.io/en/latest/installation.html#middleware
    MIDDLEWARE += ["debug_toolbar.middleware.DebugToolbarMiddleware"]
    DEBUG_TOOLBAR_PANELS = [
        "debug_toolbar.panels.versions.VersionsPanel",
        "debug_toolbar.panels.timer.TimerPanel",
        "debug_toolbar.panels.settings.SettingsPanel",
        "debug_toolbar.panels.headers.HeadersPanel",
        "debug_toolbar.panels.request.RequestPanel",
        "debug_toolbar.panels.sql.SQLPanel",
        "debug_toolbar.panels.staticfiles.StaticFilesPanel",
        "debug_toolbar.panels.templates.TemplatesPanel",
        "debug_toolbar.panels.cache.CachePanel",
        "debug_toolbar.panels.signals.SignalsPanel",
        "debug_toolbar.panels.logging.LoggingPanel",
        "debug_toolbar.panels.redirects.RedirectsPanel",
    ]
    # https://django-debug-toolbar.readthedocs.io/en/latest/configuration.html#debug-toolbar-config
    DEBUG_TOOLBAR_CONFIG = {
        "SHOW_TEMPLATE_CONTEXT": True,
        "SHOW_TOOLBAR_CALLBACK": lambda: True,
    }

# django-queryinspect
# ------------------------------------------------------------------------------
if DEVELOPMENT and env.bool("ENABLE_QUERYINSPECT", default=False):  # type: ignore[arg-type]
    # Configure django-queryinspect
    MIDDLEWARE += ["qinspect.middleware.QueryInspectMiddleware"]
    # Whether the Query Inspector should do anything (default: False)
    QUERY_INSPECT_ENABLED = True
    # Whether to log the stats via Django logging (default: True)
    QUERY_INSPECT_LOG_STATS = True
    # Whether to add stats headers (default: True)
    QUERY_INSPECT_HEADER_STATS = False
    # Whether to log duplicate queries (default: False)
    QUERY_INSPECT_LOG_QUERIES = True
    # Whether to log queries that are above an absolute limit (default: None - disabled)
    QUERY_INSPECT_ABSOLUTE_LIMIT = 100  # in milliseconds
    # Whether to log queries that are more than X standard deviations above the
    # mean query time (default: None)
    QUERY_INSPECT_STANDARD_DEVIATION_LIMIT = 2
    # Whether to include tracebacks in the logs (default: False)
    QUERY_INSPECT_LOG_TRACEBACKS = False
    # Uncomment this to filter traceback output to include only lines of our
    # app's first-party code.  I personally don't find this useful, because the
    # offending Python is sometimes actually somewhere in django core.
    # QUERY_INSPECT_TRACEBACK_ROOTS = ['/']  # noqa: ERA001

# TESTING
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/3.0/ref/settings/#std:setting-TEST_RUNNER
if TESTING:
    DEBUG = False
    # Set the root logger to only display WARNING logs and above during tests.
    logging.getLogger("").setLevel(logging.WARNING)

# LDAP Configuration
# ------------------------------------------------------------------------------
LDAP_SERVERS = {
    "default": {
        "basedn": env("LDAP_BASEDN", default="o=example,c=us"),  # type: ignore[arg-type, assignment]
        "read": {
            "url": env("LDAP_URL", default="ldap://ldap:389"),  # type: ignore[arg-type, assignment]
            "user": env("LDAP_USER", default="cn=Directory Manager"),  # type: ignore[arg-type, assignment]
            "password": env("LDAP_PASSWORD", default="password"),  # type: ignore[arg-type, assignment]
        },
        "write": {
            "url": env("LDAP_URL", default="ldap://ldap:389"),  # type: ignore[arg-type, assignment]
            "user": env("LDAP_USER", default="cn=Directory Manager"),  # type: ignore[arg-type, assignment]
            "password": env("LDAP_PASSWORD", default="password"),  # type: ignore[arg-type, assignment]
        },
    }
}



from demo.logging import logger
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.loader import MigrationLoader


class Command(BaseCommand):
    """
    Do our initial migrate and data load if necessary.

    If `settings.BOOTSTRAP_ALWAYS_MIGRATE` is `True`, always run migrations.
    """

    def db_is_fresh(self, database):
        """
        Figure out if we've never run migrations here.

        Assume that if the contenttypes.0001_initial migration has not run,
        we have a fresh database.
        """
        connection = connections[database]
        loader = MigrationLoader(connection)
        return ("contenttypes", "0001_initial") not in loader.applied_migrations

    def handle(self, **options):  # noqa: ARG002
        logger.info("migrate.start")
        if self.db_is_fresh(DEFAULT_DB_ALIAS):
            call_command("migrate")
            call_command("loaddata", "users")
        elif settings.BOOTSTRAP_ALWAYS_MIGRATE:
            call_command("migrate")
        logger.info("migrate.end")



"""
LDAP ORM type definitions.

This module provides type aliases for LDAP data structures and operations,
using Python 3.10+ type hinting conventions.
"""

DeleteModListEntry = tuple[int, str, None]
ModifyModListEntry = tuple[int, str, str]
AddModlistEntry = tuple[str, str]
ModifyDeleteModList = list[DeleteModListEntry | ModifyModListEntry | AddModlistEntry]
AddModlist = list[tuple[str, str]]
LDAPData = tuple[str, dict[str, list[bytes]]]



from typing import TYPE_CHECKING, ClassVar, cast

from rest_framework import viewsets
from rest_framework.filters import BaseFilterBackend
from rest_framework.pagination import PageNumberPagination

from demo.core.ldap.models import NSRole
from ldaporm.restframework import LdapFilterBackend, LdapOrderingFilter

from ..serializers import NSRoleSerializer

if TYPE_CHECKING:
    from ldaporm.managers import LdapManager


class NSRoleFilter(LdapFilterBackend):
    """
    Filter backend for NSRole using the abstracted LdapFilterBackend.
    """

    filter_fields: ClassVar[dict[str, dict[str, str]]] = {
        "cn": {"lookup": "iexact", "type": "string"},
        "description": {"lookup": "icontains", "type": "string"},
    }


class NSRoleViewSet(viewsets.ModelViewSet):
    """
    ViewSet for NSRole operations.
    Provides CRUD operations for NSRoles with filtering and ordering support.
    """

    serializer_class = NSRoleSerializer
    model = NSRole
    lookup_field = "cn"  # Use cn as the primary key for NSRoles
    pagination_class = PageNumberPagination  # Use standard pagination for testing
    filter_backends: ClassVar[list[BaseFilterBackend]] = [
        NSRoleFilter,
        LdapOrderingFilter,
    ]

    # Define which fields can be used for searching
    search_fields = (
        "cn",
        "description",
    )

    # Define which fields can be used for ordering
    ordering_fields = (
        "cn",
        "description",
    )

    ordering = ("cn",)

    def get_queryset(self):
        """
        Return the NSRole queryset.
        """
        return cast("LdapManager", NSRole.objects).all()



from typing import Final

from django.urls import reverse

from ldaporm.fields import (
    CharField,
    CharListField,
    DateTimeField,
    IntegerField,
)
from ldaporm.models import Model
from ldaporm.validators import EmailForwardValidator


class LDAPUser(Model):
    """
    LDAP model for POSIX users.
    """

    #: Login shell choices
    LOGIN_SHELL_CHOICES: Final[list[tuple[str, str]]] = [
        ("/bin/csh", "C Shell"),
        ("/bin/tcsh", "Tcsh"),
        ("/bin/bash", "Bash"),
        ("/bin/zsh", "Zsh"),
        ("/bin/fish", "Fish"),
    ]

    #: Employee type choices
    EMPLOYEE_TYPE_CHOICES: Final[list[tuple[str, str]]] = [
        ("employee", "Employee"),
        ("contractor", "Contractor"),
        ("intern", "Intern"),
        ("manager", "Manager"),
        ("external", "External"),
    ]

    def get_absolute_url(self) -> str:
        """
        Return the absolute URL for the user.
        """
        return reverse("core:user--detail", kwargs={"uid": self.uid})

    # Identity fields
    uid = CharField("uid", primary_key=True, max_length=50)
    full_name = CharField("Full Name", max_length=100, db_column="cn")
    last_name = CharField("Last Name", max_length=100, db_column="sn")
    first_name = CharField("First Name", max_length=100, db_column="givenName")

    # Contact information
    mail = CharListField(
        "Aliases",
        max_length=254,
        help_text="List of email addresses for this user",
        validators=[EmailForwardValidator()],
    )

    # POSIX fields
    uid_number = IntegerField("UID Number", db_column="uidNumber")
    gid_number = IntegerField("GID Number", db_column="gidNumber")
    home_directory = CharField(
        "Home Directory", max_length=200, db_column="homeDirectory"
    )
    login_shell = CharField(
        "Login Shell", max_length=50, default="/bin/bash", db_column="loginShell"
    )

    # inetOrgPerson fields
    employee_type = CharField(
        "Employee Type", max_length=50, null=True, db_column="employeeType"
    )
    employee_number = IntegerField("Employee Number", db_column="employeeNumber")
    room_number = CharField("Room Number", max_length=50, db_column="roomNumber")
    home_phone = CharField("Home Phone", max_length=50, db_column="homePhone")
    mobile = CharField("Mobile", max_length=50, db_column="mobile")

    # 389-ds fields
    created_at = DateTimeField(
        "Created At",
        null=True,
        blank=True,
        default=None,
        db_column="createTimestamp",
        editable=False,
    )
    created_by = CharField(
        "Created By",
        null=True,
        blank=True,
        default=None,
        db_column="creatorsName",
        editable=False,
    )
    updated_at = DateTimeField(
        "Updated At",
        null=True,
        blank=True,
        default=None,
        db_column="modifyTimestamp",
        editable=False,
    )
    updated_by = CharField(
        "Updated By",
        null=True,
        blank=True,
        default=None,
        db_column="modifiersName",
        editable=False,
    )

    nsroledn = CharListField(
        "nsroledn",
        max_length=255,
        help_text="List of NSRole DNs",
    )
    nsrole = CharListField(
        "nsrole",
        blank=True,
        default=None,
        editable=False,
        help_text="List of all nsroles, managed, search and nested roles",
    )

    class Meta:
        ldap_server: str = "default"
        basedn: str = "ou=people,o=example,c=us"
        objectclass: str = "posixAccount"
        extra_objectclasses: list[str] = [  # noqa: RUF012
            "top",
            "inetOrgPerson",
        ]
        verbose_name: str = "LDAP User"
        verbose_name_plural: str = "LDAP Users"
        ordering: list[str] = ["uid"]  # noqa: RUF012
        password_attribute: str = "userPassword"  # noqa: S105

    def get_full_name(self) -> str:
        """Return the user's full name."""
        return f"{self.first_name} {self.last_name}"

    def __str__(self) -> str:
        return f"{self.full_name} (uid: {self.uid})"

    def __repr__(self) -> str:
        return f"<LDAPUser: {self.uid}>"


class LDAPGroup(Model):
    """LDAP model for POSIX groups."""

    cn = CharField(
        "Common Name", primary_key=True, max_length=50, help_text="Common Name"
    )
    gid_number = IntegerField(
        "GID Number", db_column="gidNumber", help_text="GID Number"
    )
    member_uids = CharListField(
        "Member UID",
        max_length=50,
        db_column="memberUid",
        help_text="List of UIDs of group members",
    )
    description = CharField(
        "Description", max_length=200, blank=True, help_text="Description of the group"
    )

    class Meta:
        ldap_server = "default"
        basedn = "ou=groups,o=example,c=us"
        objectclass = "posixGroup"
        extra_objectclasses = [  # noqa: RUF012
            "top",
        ]
        verbose_name = "LDAP Group"
        verbose_name_plural = "LDAP Groups"
        ordering = ["cn"]  # noqa: RUF012

    def get_member_count(self) -> int:
        """Return the number of members in the group."""
        return len(self.member_uids) if self.member_uids else 0  # type: ignore[arg-type]

    def __str__(self) -> str:
        return f"{self.cn} (gid: {self.gid_number})"

    def __repr__(self) -> str:
        return f"<LDAPGroup: {self.cn}>"


class NSRole(Model):
    """LDAP model for NSRoles."""

    cn = CharField(
        "Common Name", primary_key=True, max_length=50, help_text="Common Name"
    )
    description = CharField(
        "Description", max_length=200, blank=True, help_text="Description of the role"
    )

    class Meta:
        ldap_server = "default"
        basedn = "ou=roles,o=example,c=us"
        objectclass = "ldapsubentry"
        extra_objectclasses = [  # noqa: RUF012
            "top",
            "nsroledefinition",
            "nssimpleroledefinition",
            "nsmanagedroledefinition",
        ]
        verbose_name = "Role"
        verbose_name_plural = "Roles"
        ordering = ["cn"]  # noqa: RUF012



import base64
import datetime
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, ClassVar, cast

from rest_framework import Request, View, serializers
from rest_framework.filters import BaseFilterBackend
from rest_framework.pagination import BasePagination
from rest_framework.response import Response
from rest_framework.reverse import reverse

from ldaporm import fields
from ldaporm.managers import F
from ldaporm.models import Model as LdapModel

if TYPE_CHECKING:
    from ldaporm.managers import Manager as LdapManager
    from ldaporm.options import Options as LdapOptions


class BinaryField(serializers.Field):
    """
    A custom serializer field for handling binary data.

    This field handles the conversion between base64-encoded strings in JSON
    and Python bytes objects for LDAP binary attributes.
    """

    def to_internal_value(self, data):
        """
        Convert base64-encoded string to bytes.

        Args:
            data: Base64-encoded string data.

        Returns:
            Decoded bytes data.

        Raises:
            ValidationError: If the data cannot be decoded.

        """
        if data is None:
            return None

        if isinstance(data, bytes):
            return data

        if isinstance(data, str):
            try:
                return base64.b64decode(data.encode("utf-8"))
            except Exception as e:
                msg = f"Invalid base64 data: {e}"
                raise serializers.ValidationError(msg) from e

        msg = "Binary field must be a base64-encoded string or bytes"
        raise serializers.ValidationError(msg)

    def to_representation(self, value):
        """
        Convert bytes to base64-encoded string.

        Args:
            value: Bytes data.

        Returns:
            Base64-encoded string.

        """
        if value is None:
            return None

        if isinstance(value, bytes):
            return base64.b64encode(value).decode("utf-8")

        return value


class LdapModelSerializer(serializers.Serializer):
    """
    A DRF Serializer for ldaporm.Model subclasses.
    Automatically introspects ldaporm fields and provides create/update methods.
    """

    def __init__(self, *args, **kwargs):
        # Dynamically add fields from the ldaporm model
        model_class = self.Meta.model
        assert issubclass(model_class, LdapModel), (  # noqa: S101
            "Meta.model must be a subclass of ldaporm.Model"
        )
        super().__init__(*args, **kwargs)
        for field in model_class._meta.fields:
            if field.name in self.fields:
                continue  # Already defined explicitly
            drf_field = self._get_drf_field(field)
            self.fields[field.name] = drf_field

    def _get_drf_field(self, ldap_field: fields.Field) -> serializers.Field:  # noqa: PLR0912
        # Determine required/optional logic
        # Primary key fields should always be required
        is_pk = getattr(ldap_field, "primary_key", False)
        if is_pk:
            is_required = True
        else:
            # For non-primary key fields, check blank and null attributes
            is_blank = getattr(ldap_field, "blank", False)
            allow_null = getattr(ldap_field, "null", False)

            # A field is required if it's not blank and not null
            is_required = not is_blank and not allow_null

        allow_null = getattr(ldap_field, "null", False)

        # Map ldaporm field types to DRF fields (basic mapping, extend as needed)
        if isinstance(ldap_field, fields.CharListField):
            field = serializers.ListField(
                child=serializers.CharField(),
                required=is_required,
                allow_null=allow_null,
            )
        elif hasattr(ldap_field, "choices") and ldap_field.choices:
            field = serializers.ChoiceField(
                choices=ldap_field.choices, required=is_required, allow_null=allow_null
            )
        elif isinstance(ldap_field, (fields.EmailField, fields.EmailForwardField)):
            field = serializers.EmailField(required=is_required, allow_null=allow_null)
        elif isinstance(ldap_field, fields.CaseInsensitiveSHA1Field):
            # Read-only field for hashed values
            field = serializers.CharField(
                read_only=True,
                style={"input_type": "password"},
            )
        elif isinstance(ldap_field, (fields.LDAPPasswordField, fields.ADPasswordField)):
            # Password fields should be read-only and styled as password
            field = serializers.CharField(
                read_only=True,
                style={"input_type": "password"},
            )
        elif isinstance(ldap_field, fields.CharField):
            field = serializers.CharField(
                required=is_required,
                allow_blank=getattr(ldap_field, "blank", False),
                allow_null=allow_null,
            )
        elif isinstance(ldap_field, fields.IntegerField):
            field = serializers.IntegerField(
                required=is_required, allow_null=allow_null
            )
        elif isinstance(ldap_field, (fields.BooleanField, fields.AllCapsBooleanField)):
            field = serializers.BooleanField(
                required=is_required, allow_null=allow_null
            )
        elif isinstance(
            ldap_field, (fields.DateTimeField, fields.ActiveDirectoryTimestampField)
        ):
            field = serializers.DateTimeField(
                required=is_required, allow_null=allow_null
            )
        elif isinstance(ldap_field, fields.DateField):
            field = serializers.DateField(required=is_required, allow_null=allow_null)
        elif isinstance(ldap_field, fields.BinaryField):
            field = BinaryField(
                required=is_required,
                allow_null=allow_null,
            )
        else:
            field = serializers.CharField(
                required=is_required,
                allow_blank=getattr(ldap_field, "blank", False),
                allow_null=allow_null,
            )
        return field

    def create(self, validated_data):
        model_class = self.Meta.model
        instance = model_class(**validated_data)
        instance.save()
        return instance

    def update(self, instance, validated_data):
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    def to_representation(self, instance):
        # Convert model instance to dict for DRF
        ret = {}
        for field in instance._meta.fields:
            value = getattr(instance, field.name)
            # Convert date objects to ISO format strings for JSON serialization
            if isinstance(value, datetime.date):
                ret[field.name] = value.isoformat()
            # Convert binary data to base64 for JSON serialization
            elif isinstance(field, fields.BinaryField) and value is not None:
                ret[field.name] = base64.b64encode(value).decode("utf-8")
            else:
                ret[field.name] = value
        if hasattr(instance, "dn"):
            ret["dn"] = instance.dn
        return ret

    class Meta:
        model = (
            None  # Set this to your ldaporm.Model subclass in your serializer subclass
        )


class HyperlinkedModelSerializer(LdapModelSerializer):
    """
    A hyperlinked serializer for ldaporm.Model subclasses.

    This serializer provides URL-based identification and hyperlinked relationships
    for LDAP ORM models, similar to Django REST Framework's HyperlinkedModelSerializer.

    The serializer automatically includes a 'url' field that points to the detail
    view for each instance, and can handle hyperlinked relationships to other
    LDAP ORM models.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Add the url field if not already present
        if "url" not in self.fields:
            # Get extra kwargs for the url field if specified
            extra_kwargs = getattr(self.Meta, "extra_kwargs", {})
            url_kwargs = extra_kwargs.get("url", {})

            self.fields["url"] = serializers.HyperlinkedIdentityField(
                view_name=url_kwargs.get("view_name", self._get_detail_view_name()),
                lookup_field=url_kwargs.get("lookup_field", self._get_lookup_field()),
            )

    def _get_detail_view_name(self) -> str:
        """
        Get the view name for the detail endpoint.

        Returns:
            The view name for the detail endpoint, constructed from the model name.

        """
        meta = cast("LdapOptions", cast("LdapModel", self.Meta.model)._meta)
        name = cast("str", meta.object_name)
        return f"{name.lower()}-detail"

    def _get_lookup_field(self) -> str:
        """
        Get the lookup field for URL construction.

        For LDAP ORM models, we typically use 'dn' as the lookup field
        since it's the unique identifier in LDAP.

        Returns:
            The lookup field name, defaults to 'dn'.

        """
        return getattr(self.Meta, "lookup_field", "dn")

    def _get_drf_field(self, ldap_field: fields.Field) -> serializers.Field:
        """
        Get DRF field with support for hyperlinked relationships.

        Args:
            ldap_field: The LDAP ORM field to convert.

        Returns:
            A DRF serializer field.

        """
        # Check if this field should be a hyperlinked relationship
        if self._is_relationship_field(ldap_field):
            return self._get_hyperlinked_relationship_field(ldap_field)
        # Use parent's field mapping for regular fields
        return super()._get_drf_field(ldap_field)

    def _is_relationship_field(self, ldap_field: fields.Field) -> bool:
        """
        Check if a field should be treated as a relationship field.

        Args:
            ldap_field: The LDAP ORM field to check.

        Returns:
            True if the field should be a relationship, False otherwise.

        """
        # Check if field name suggests a relationship (ends with '_dn' or '_id')
        if ldap_field.name and ldap_field.name.endswith(("_dn", "_id")):
            return True

        # Check if field has a related model specified in Meta
        if hasattr(self.Meta, "relationship_fields"):
            if ldap_field.name in self.Meta.relationship_fields:  # type: ignore[operator]
                return True

        # Check if field is in relationship_models mapping
        if hasattr(self.Meta, "relationship_models"):
            if ldap_field.name in self.Meta.relationship_models:  # type: ignore[operator]
                return True

        return False

    def _get_hyperlinked_relationship_field(
        self, ldap_field: fields.Field
    ) -> serializers.Field:
        """
        Create a hyperlinked relationship field.

        Args:
            ldap_field: The LDAP ORM field representing the relationship.

        Returns:
            A DRF HyperlinkedRelatedField.

        """
        # Determine the related model and view name
        related_model = cast(
            "type[LdapModel] | None", self._get_related_model(ldap_field)
        )
        if related_model:
            meta = cast("LdapOptions", related_model._meta)
            name = cast("str", meta.object_name)
            default_view_name = f"{name.lower()}-detail"

            # Get extra kwargs for this field if specified
            extra_kwargs = getattr(self.Meta, "extra_kwargs", {})
            field_kwargs = extra_kwargs.get(cast("str", ldap_field.name), {})

            return serializers.HyperlinkedRelatedField(
                view_name=field_kwargs.get("view_name", default_view_name),
                lookup_field=field_kwargs.get("lookup_field", "dn"),
                queryset=cast("LdapManager", related_model.objects),
                required=not ldap_field.blank,
                allow_null=ldap_field.null,
            )

        # Fallback to regular field if we can't determine the relationship
        return super()._get_drf_field(ldap_field)

    def _get_related_model(self, ldap_field: fields.Field) -> type[LdapModel] | None:
        """
        Get the related model for a relationship field.

        Args:
            ldap_field: The LDAP ORM field representing the relationship.

        Returns:
            The related model class or None if not found.

        """
        # Check if relationship mapping is defined in Meta
        if hasattr(self.Meta, "relationship_models"):
            try:
                return cast(
                    "type[LdapModel]",
                    self.Meta.relationship_models[cast("str", ldap_field.name)],
                )
            except KeyError:
                # Return None instead of raising KeyError
                return None

        # Return None if no relationship_models configured
        return None

    def to_representation(self, instance):
        """
        Convert model instance to dict with hyperlinked URLs.

        Args:
            instance: The LDAP ORM model instance.

        Returns:
            A dictionary representation with hyperlinked URLs.

        """
        ret = super().to_representation(instance)

        # Add hyperlinked relationships
        for field in instance._meta.fields:
            if self._is_relationship_field(field):
                related_value = getattr(instance, field.name)
                if related_value:
                    # Try to get the related object and create a URL
                    try:
                        related_model = self._get_related_model(field)
                        if related_model:
                            # This is a simplified approach - you might need to
                            # implement proper LDAP querying here
                            related_obj = related_model.objects.get(dn=related_value)
                            view_name = (
                                f"{related_model._meta.object_name.lower()}-detail"
                            )
                            ret[field.name] = reverse(
                                view_name,
                                kwargs={"dn": related_obj.dn},
                                request=self.context.get("request"),
                                format=self.context.get("format"),
                            )
                    except (related_model.DoesNotExist, AttributeError):
                        # If we can't resolve the relationship, just include the DN
                        ret[field.name] = related_value

        return ret

    class Meta:
        #: The model to serialize.
        model: ClassVar[type[LdapModel] | None] = (
            None  # Set this to your ldaporm.Model subclass
        )
        #: The field to use for URL lookups.
        lookup_field: ClassVar[str] = "dn"  # The field to use for URL lookups
        #: A list of relationship fields.
        relationship_fields: ClassVar[Iterable[str]] = ()
        #: A dictionary of relationship fields to their related models.
        relationship_models: ClassVar[dict[str, LdapModel]] = {}
        #: Extra keyword arguments for field configuration.
        extra_kwargs: ClassVar[dict[str, dict[str, str]]] = {}


class LdapCursorPagination(BasePagination):
    """
    LDAP cursor-based pagination for REST framework views.

    This pagination class uses LDAP's SimplePagedResultsControl to provide
    efficient server-side paging. It uses base64-encoded cookies as cursors
    to maintain pagination state.

    Example usage:

        .. code-block:: python

            from rest_framework import viewsets
            from ldaporm.restframework import LdapCursorPagination

            from .models import User

            class UserSerializer(LdapModelSerializer):
                class Meta:
                    model = User
                    fields = '__all__'

            class UserViewSet(viewsets.ModelViewSet):
                pagination_class = LdapCursorPagination
                serializer_class = UserSerializer

                def get_queryset(self):
                    return User.objects.all()

    """

    #: The default page size.
    page_size = 100
    #: The query parameter name for the page size.
    page_size_query_param = "page_size"
    #: The maximum page size.
    max_page_size = 1000
    #: The query parameter name for the cursor.
    cursor_query_param = "next_token"

    def paginate_queryset(
        self,
        queryset: F,
        request: Request,
        view: View | None = None,  # noqa: ARG002
    ) -> list[LdapModel]:
        """
        Paginate the queryset using LDAP paging.

        Args:
            queryset: The LDAP ORM queryset to paginate.
            request: The HTTP request.
            view: The view instance.

        Returns:
            A PagedResultSet containing the current page of results.

        """
        # Get page size from request
        page_size = self.get_page_size(request)

        # Get cursor from request
        cursor = request.query_params.get(self.cursor_query_param, "")

        # Decode cursor if provided
        cookie = ""
        if cursor:
            try:
                cookie = base64.b64decode(cursor.encode()).decode()
            except Exception:  # noqa: BLE001
                # Invalid cursor, start from beginning
                cookie = ""

        # Perform paged search
        paged_results = queryset.page(page_size=page_size, cookie=cookie)

        # Store pagination info for response
        self.paged_results = paged_results
        self.request = request

        return paged_results.results

    def get_paginated_response(self, data: list[Any]) -> Response:
        """
        Return a paginated response with cursor-based navigation.

        Args:
            data: The serialized data for the current page.

        Returns:
            A Response with pagination metadata.

        """
        response_data = {
            "results": data,
            "has_more": self.paged_results.has_more,
        }

        # Add next cursor if there are more results
        if self.paged_results.has_more and self.paged_results.next_cookie:
            next_cursor = base64.b64encode(
                self.paged_results.next_cookie.encode()
            ).decode()
            response_data["next"] = self._get_next_url(next_cursor)

        return Response(response_data)

    def get_page_size(self, request: Request) -> int:
        """
        Get the page size from the request.

        Args:
            request: The HTTP request.

        Returns:
            The page size to use.

        """
        if self.page_size_query_param:
            try:
                # Handle both Django test requests and DRF requests
                if hasattr(request, "query_params"):
                    # DRF request
                    page_size = int(
                        request.query_params.get(
                            self.page_size_query_param, self.page_size
                        )
                    )
                else:
                    # Django test request
                    page_size = int(
                        request.GET.get(self.page_size_query_param, self.page_size)
                    )
                return min(page_size, self.max_page_size)
            except (KeyError, ValueError):
                pass
        return self.page_size

    def _get_next_url(self, cursor: str) -> str:
        """
        Generate the URL for the next page.

        Args:
            cursor: The base64-encoded cursor for the next page.

        Returns:
            The URL for the next page.

        """
        from urllib.parse import parse_qs, urlencode, urlparse

        # Parse the current URL
        parsed_url = urlparse(self.request.build_absolute_uri())
        query_params = parse_qs(parsed_url.query)

        # Update or add the cursor parameter
        query_params[self.cursor_query_param] = [cursor]

        # Preserve all other query parameters (including page_size)
        # Remove the cursor parameter from the original URL to avoid duplication
        if self.cursor_query_param in query_params:
            # Ensure we have the latest cursor value
            pass

        # Rebuild the query string
        new_query = urlencode(query_params, doseq=True)

        # Reconstruct the URL
        new_url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"
        if new_query:
            new_url = f"{new_url}?{new_query}"

        return new_url


class LdapOrderingFilter(BaseFilterBackend):
    """
    A filter backend that provides ordering for LDAP ORM models.

    This filter leverages LDAP ORM's built-in ordering capabilities, which use
    server-side sorting when available and fall back to client-side sorting
    when the LDAP server doesn't support server-side sorting.

    The filter supports:

    - Single and multiple field ordering
    - Ascending and descending ordering (using '-' prefix)
    - Field validation against the LDAP model's available fields
    - Fallback to model's default ordering when no ordering is specified

    Example usage:

        .. code-block:: python

            from ldaporm.restframework import LdapOrderingFilter
            from rest_framework import viewsets

            from .models import User

            class UserViewSet(viewsets.ModelViewSet):
                filter_backends = [LdapOrderingFilter]
                ordering_fields = ['uid', 'cn', 'mail', 'created']
                ordering = ['uid']  # Default ordering

                def get_queryset(self):
                    return User.objects.all()

    Query parameters:
        ``?ordering=uid,-cn,mail``  # Order by uid ascending, cn descending,
        mail ascending
    """

    #: The query parameter name for ordering
    ordering_param = "ordering"
    #: The fields that can be used for ordering (None means all fields)
    ordering_fields = None
    #: The default ordering to use when no ordering is specified
    ordering = None

    def get_ordering(self, request, queryset, view):
        """
        Get the ordering from the request and validate it.

        Args:
            request: The HTTP request
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            A list of ordering fields (with '-' prefix for descending)

        """
        params = request.query_params.get(self.ordering_param)
        if not params:
            return self.get_default_ordering(view, queryset)

        fields = [param.strip() for param in params.split(",")]
        ordering = []

        for field in fields:
            if not field:
                continue

            # Check if field is valid
            if self._is_valid_field(field, queryset, view):
                ordering.append(field)
            # If field is invalid, just ignore it instead of raising an error

        return ordering

    def filter_queryset(self, request, queryset, view):
        """
        Apply ordering to the LDAP ORM queryset.

        This method leverages LDAP ORM's built-in ordering capabilities,
        which automatically use server-side sorting when available and
        fall back to client-side sorting when needed.

        Args:
            request: The HTTP request
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            The ordered LDAP ORM queryset

        """
        ordering = self.get_ordering(request, queryset, view)
        if ordering:
            return queryset.order_by(*ordering)
        return queryset

    def get_default_ordering(self, view, queryset=None):
        """
        Get the default ordering from the view or model.

        Args:
            view: The view instance
            queryset: The LDAP ORM queryset (optional)

        Returns:
            A list of default ordering fields

        """
        # Check view's ordering attribute first
        if hasattr(view, "ordering") and view.ordering is not None:
            return view.ordering

        # Check filter's default ordering
        if self.ordering is not None:
            return self.ordering

        # Use provided queryset or get from view
        if queryset is None and hasattr(view, "get_queryset"):
            queryset = view.get_queryset()

        # Fall back to model's default ordering
        if queryset and hasattr(queryset, "model") and hasattr(queryset.model, "_meta"):
            return getattr(queryset.model._meta, "ordering", [])

        return []

    def _is_valid_field(self, field, queryset, view):
        """
        Check if a field is valid for ordering.

        Args:
            field: The field name (with or without '-' prefix)
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            True if the field is valid, False otherwise

        """
        # Remove '-' prefix for validation
        field_name = field.lstrip("-")

        # Check against view's ordering_fields if specified
        ordering_fields = self._get_ordering_fields(view)
        if ordering_fields is not None:
            return field_name in ordering_fields

        # Check against model's available fields
        if hasattr(queryset, "model") and hasattr(queryset.model, "_meta"):
            model_fields = queryset.model._meta.fields
            return field_name in [f.name for f in model_fields]

        return True

    def _get_ordering_fields(self, view):
        """
        Get the allowed ordering fields from the view.

        Args:
            view: The view instance

        Returns:
            A list of allowed ordering fields, or None if all fields are allowed

        """
        # Check view's ordering_fields attribute
        if hasattr(view, "ordering_fields") and view.ordering_fields is not None:
            return view.ordering_fields

        # Check filter's ordering_fields attribute
        return self.ordering_fields

    def _get_available_fields(self, queryset, view):
        """
        Get the list of available fields for ordering.

        Args:
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            A list of available field names

        """
        ordering_fields = self._get_ordering_fields(view)
        if ordering_fields is not None:
            return ordering_fields

        if hasattr(queryset, "model") and hasattr(queryset.model, "_meta"):
            return [f.name for f in queryset.model._meta.fields]

        return []

    def get_schema_operation_parameters(self, view):
        """
        Get the schema operation parameters for OpenAPI documentation.

        Args:
            view: The view instance

        Returns:
            A list of parameter definitions for OpenAPI

        """
        available_fields = []
        if hasattr(view, "get_queryset"):
            queryset = view.get_queryset()
            available_fields = self._get_available_fields(queryset, view)

        return [
            {
                "name": self.ordering_param,
                "required": False,
                "in": "query",
                "description": (
                    "Ordering field(s). Use '-' prefix for descending order. "
                    f"Available fields: {', '.join(available_fields)}"
                ),
                "schema": {
                    "type": "string",
                    "example": "uid,-cn,mail",
                },
            }
        ]


class LdapFilterBackend(BaseFilterBackend):
    """
    Base class for LDAP model filtering backends.

    This class provides a foundation for creating custom filter backends
    that work with LDAP ORM models. It handles common filtering patterns
    and provides a consistent interface for LDAP-specific filtering.

    Subclasses should override:

    - filter_fields: Define the fields that can be filtered.  There should
      be a dictionary of field names to their configurations.  The
      configurations should include:

      - lookup: The lookup type to use for the field.  This should be one of
        the following:

        - iexact: case-insensitive exact match
        - icontains: case-insensitive contains
        - istartswith: starts with
        - iendswith: ends with
        - gt: greater than (numeric)
        - gte: greater than or equal to (numeric)
        - lt: less than (numeric)
        - lte: less than or equal to (numeric)

      - type: The field type to use for the field.  This should be one of the
        following:

        - string: string
        - integer: integer
        - boolean: boolean
        - binary: binary

    Example usage:

        .. code-block:: python

            from rest_framework import viewsets
            from ldaporm.restframework import LdapFilterBackend
            from .models import User

            class UserFilterBackend(LdapFilterBackend):
                filter_fields = {
                    'uid': {'lookup': 'iexact', 'type': 'string'},
                    'mail': {'lookup': 'icontains', 'type': 'string'},
                    'employee_number': {'lookup': 'iexact', 'type': 'integer'},
                }

            class UserViewSet(viewsets.ModelViewSet):
                model = User
                filter_backends = [UserFilterBackend]

    """

    #: Dictionary defining filterable fields and their configurations
    filter_fields: ClassVar[dict[str, dict[str, str]]] = {}

    def filter_queryset(self, request, queryset, view):
        """
        Filter the queryset based on query parameters.

        Args:
            request: The HTTP request
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            The filtered LDAP ORM queryset

        """
        return self.get_filter_queryset(request, queryset, view)

    def get_filter_queryset(self, request: Request, queryset: F, view: View) -> F:  # noqa: ARG002
        """
        Apply filters to the queryset. Override this method in subclasses.  This
        method is called by :py:meth:`filter_queryset` and should return the
        filtered queryset.

        Args:
            request: The HTTP request
            queryset: The LDAP ORM queryset
            view: The view instance

        Returns:
            The filtered LDAP ORM queryset

        """
        return self.apply_filters(request, queryset)

    def apply_filters(self, request: Request, queryset: F) -> F:
        """
        Apply filters based on the filter_fields configuration.  This method is
        called by :py:meth:`get_filter_queryset` and should return the filtered
        queryset.

        Args:
            request: The HTTP request
            queryset: The LDAP ORM queryset

        Returns:
            The filtered LDAP ORM queryset

        """
        for field_name, field_config in self.filter_fields.items():
            value = request.query_params.get(field_name)
            if value is not None:
                lookup = field_config.get("lookup", "iexact")
                field_type = field_config.get("type", "string")

                # Convert value based on field type
                if field_type == "integer":
                    try:
                        value = int(value)
                    except (ValueError, TypeError):
                        continue  # Skip invalid integer values
                elif field_type == "boolean":
                    if value.lower() in ("true", "1", "yes", "on"):
                        value = True
                    elif value.lower() in ("false", "0", "no", "off"):
                        value = False
                    else:
                        continue  # Skip invalid boolean values
                elif field_type == "binary":
                    try:
                        # Decode base64-encoded binary data
                        value = base64.b64decode(value.encode("utf-8"))
                    except Exception:  # noqa: BLE001, S112
                        continue  # Skip invalid base64 values

                # Apply the filter
                filter_kwargs = {f"{field_name}__{lookup}": value}
                queryset = queryset.filter(**filter_kwargs)

        return queryset

    def get_schema_operation_parameters(self, view: View) -> list[dict[str, Any]]:  # noqa: ARG002
        """
        Return schema operation parameters for OpenAPI documentation.

        Args:
            view: The view instance

        Returns:
            A list of parameter definitions for OpenAPI

        """
        parameters = []
        for field_name, field_config in self.filter_fields.items():
            field_type = field_config.get("type", "string")
            lookup = field_config.get("lookup", "iexact")

            # Generate description based on lookup type
            lookup_descriptions = {
                "iexact": "case-insensitive exact match",
                "icontains": "case-insensitive contains",
                "exact": "exact match",
                "contains": "contains",
                "startswith": "starts with",
                "endswith": "ends with",
            }
            lookup_desc = lookup_descriptions.get(lookup, lookup)

            # Map field type to OpenAPI schema type
            schema_type = {
                "string": "string",
                "integer": "integer",
                "boolean": "boolean",
                "float": "number",
                "binary": "string",  # Binary data is transmitted as base64 string
            }.get(field_type, "string")

            parameters.append(
                {
                    "name": field_name,
                    "required": False,
                    "in": "query",
                    "description": f"Filter by {field_name} ({lookup_desc})",
                    "schema": {"type": schema_type},
                }
            )

        return parameters
